# stdlib
from hashlib import sha256
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Union

# third party
from packaging import version
from pydantic import validator
from typing_extensions import Self
import yaml

# relative
from ..serde.serializable import serializable
from ..types.base import SyftBaseModel

PYTHON_DEFAULT_VER = "3.11"
PYTHON_MIN_VER = version.parse("3.10")
PYTHON_MAX_VER = version.parse("3.12")


def _malformed_python_package_error_msg(pkg: str, name: str = "package_name") -> str:
    return f'You must pin the package to an exact version. Got "{pkg}" expected "{name}==x.y.z"'


class CustomBuildConfig(SyftBaseModel):
    gpu: bool = False
    # python_version: str = PYTHON_DEFAULT_VER
    python_packages: List[str] = []
    system_packages: List[str] = []
    custom_cmds: List[str] = []

    # @validator("python_version")
    # def validate_python_version(cls, ver: str) -> str:
    #     parsed_ver = version.parse(ver)

    #     # TODO: Check if Wolfi OS/apk supports minor version of python
    #     if parsed_ver.micro != 0:
    #         raise ValueError("Provide only major.minor version of python")

    #     if PYTHON_MIN_VER <= parsed_ver < PYTHON_MAX_VER:
    #         return ver
    #     else:
    #         raise ValueError(
    #             f"Python version must be between {PYTHON_MIN_VER} and {PYTHON_MAX_VER}"
    #         )

    @validator("python_packages")
    def validate_python_packages(cls, pkgs: List[str]) -> List[str]:
        for pkg in pkgs:
            ver_parts = ()
            name_ver = pkg.split("==")
            if len(name_ver) != 2:
                raise ValueError(_malformed_python_package_error_msg(pkg))

            name, ver = name_ver

            if ver:
                ver_parts = ver.split(".")

            if not ver or len(ver_parts) <= 2:
                raise ValueError(_malformed_python_package_error_msg(pkg, name))

        return pkgs

    def merged_python_pkgs(self, sep=" ") -> str:
        return sep.join(self.python_packages)

    def merged_system_pkgs(self, sep=" ") -> str:
        return sep.join(self.system_packages)

    def merged_custom_cmds(self, sep=";") -> str:
        return sep.join(self.custom_cmds)


class WorkerConfig(SyftBaseModel):
    pass


@serializable()
class CustomWorkerConfig(WorkerConfig):
    build: CustomBuildConfig
    version: str = "1"

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> Self:
        return cls(**config)

    @classmethod
    def from_str(cls, content: str) -> Self:
        config = yaml.safe_load(content)
        return cls.from_dict(config)

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> Self:
        with open(path) as f:
            config = yaml.safe_load(f)
            return cls.from_dict(config)

    def get_signature(self) -> str:
        return sha256(self.json(sort_keys=True).encode()).hexdigest()


@serializable()
class DockerWorkerConfig(WorkerConfig):
    dockerfile: str

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> Self:
        with open(path) as f:
            return cls(dockerfile=f.read())

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, DockerWorkerConfig):
            return False
        return self.dockerfile == __value.dockerfile

    def __hash__(self) -> int:
        return hash(self.dockerfile)



# relative
from ...serde.serializable import serializable
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from .worker_image import SyftWorkerImage


@serializable()
class SyftWorkerImageStash(BaseUIDStoreStash):
    object_type = SyftWorkerImage
    settings: PartitionSettings = PartitionSettings(
        name=SyftWorkerImage.__canonical_name__,
        object_type=SyftWorkerImage,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Type
from typing import Union

# third party
from pymongo import ASCENDING
from pymongo.collection import Collection as MongoCollection
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ..node.credentials import SyftVerifyKey
from ..serde.deserialize import _deserialize
from ..serde.serializable import serializable
from ..serde.serialize import _serialize
from ..service.action.action_permissions import ActionObjectEXECUTE
from ..service.action.action_permissions import ActionObjectOWNER
from ..service.action.action_permissions import ActionObjectPermission
from ..service.action.action_permissions import ActionObjectREAD
from ..service.action.action_permissions import ActionObjectWRITE
from ..service.action.action_permissions import ActionPermission
from ..service.context import AuthedServiceContext
from ..service.response import SyftSuccess
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import StorableObjectType
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.transforms import TransformContext
from ..types.transforms import transform
from ..types.transforms import transform_method
from ..types.uid import UID
from .document_store import DocumentStore
from .document_store import PartitionKey
from .document_store import PartitionSettings
from .document_store import QueryKey
from .document_store import QueryKeys
from .document_store import StoreConfig
from .document_store import StorePartition
from .kv_document_store import KeyValueBackingStore
from .locks import LockingConfig
from .locks import NoLockingConfig
from .mongo_client import MongoClient
from .mongo_client import MongoStoreClientConfig


@serializable()
class MongoDict(SyftBaseObject):
    __canonical_name__ = "MongoDict"
    __version__ = SYFT_OBJECT_VERSION_1

    keys: List[Any]
    values: List[Any]

    @property
    def dict(self) -> Dict[Any, Any]:
        return dict(zip(self.keys, self.values))

    @staticmethod
    def from_dict(input: Dict[Any, Any]) -> Self:
        return MongoDict(keys=list(input.keys()), values=list(input.values()))

    def __repr__(self):
        return self.dict.__repr__()


class MongoBsonObject(StorableObjectType, dict):
    pass


def _repr_debug_(value: Any) -> str:
    if hasattr(value, "_repr_debug_"):
        return value._repr_debug_()
    return repr(value)


def to_mongo(context: TransformContext) -> TransformContext:
    output = {}
    unique_keys_dict = context.obj._syft_unique_keys_dict()
    search_keys_dict = context.obj._syft_searchable_keys_dict()
    all_dict = unique_keys_dict
    all_dict.update(search_keys_dict)
    for k in all_dict:
        value = getattr(context.obj, k, "")
        # if the value is a method, store its value
        if callable(value):
            output[k] = value()
        else:
            output[k] = value

    if "id" in context.output:
        output["_id"] = context.output["id"]
    output["__canonical_name__"] = context.obj.__canonical_name__
    output["__version__"] = context.obj.__version__
    output["__blob__"] = _serialize(context.obj, to_bytes=True)
    output["__arepr__"] = _repr_debug_(context.obj)  # a comes first in alphabet
    context.output = output
    return context


@transform(SyftObject, MongoBsonObject)
def syft_obj_to_mongo():
    return [to_mongo]


@transform_method(MongoBsonObject, SyftObject)
def from_mongo(
    storage_obj: Dict, context: Optional[TransformContext] = None
) -> SyftObject:
    return _deserialize(storage_obj["__blob__"], from_bytes=True)


@serializable(attrs=["storage_type"])
class MongoStorePartition(StorePartition):
    """Mongo StorePartition

    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for partitioning and indexing.
        `store_config`: MongoStoreConfig
            Mongo specific configuration
    """

    storage_type: StorableObjectType = MongoBsonObject

    def init_store(self) -> Result[Ok, Err]:
        store_status = super().init_store()
        if store_status.is_err():
            return store_status

        client = MongoClient(config=self.store_config.client_config)

        collection_status = client.with_collection(
            collection_settings=self.settings, store_config=self.store_config
        )
        if collection_status.is_err():
            return collection_status

        collection_permissions_status = client.with_collection_permissions(
            collection_settings=self.settings, store_config=self.store_config
        )
        if collection_permissions_status.is_err():
            return collection_permissions_status

        self._collection = collection_status.ok()
        self._permissions = collection_permissions_status.ok()

        return self._create_update_index()

    # Potentially thread-unsafe methods.
    #
    # CAUTION:
    #       * Don't use self.lock here.
    #       * Do not call the public thread-safe methods here(with locking).
    # These methods are called from the public thread-safe API, and will hang the process.

    def _create_update_index(self) -> Result[Ok, Err]:
        """Create or update mongo database indexes"""
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        def check_index_keys(current_keys, new_index_keys):
            current_keys.sort()
            new_index_keys.sort()
            return current_keys == new_index_keys

        syft_obj = self.settings.object_type

        unique_attrs = getattr(syft_obj, "__attr_unique__", [])
        object_name = syft_obj.__canonical_name__

        new_index_keys = [(attr, ASCENDING) for attr in unique_attrs]

        try:
            current_indexes = collection.index_information()
        except BaseException as e:
            return Err(str(e))
        index_name = f"{object_name}_index_name"

        current_index_keys = current_indexes.get(index_name, None)

        if current_index_keys is not None:
            keys_same = check_index_keys(current_index_keys["key"], new_index_keys)
            if keys_same:
                return Ok()

            # Drop current index, since incompatible with current object
            try:
                collection.drop_index(index_or_name=index_name)
            except Exception:
                return Err(
                    f"Failed to drop index for object: {object_name} with index keys: {current_index_keys}"
                )

        # If no new indexes, then skip index creation
        if len(new_index_keys) == 0:
            return Ok()

        try:
            collection.create_index(new_index_keys, unique=True, name=index_name)
        except Exception:
            return Err(
                f"Failed to create index for {object_name} with index keys: {new_index_keys}"
            )

        return Ok()

    @property
    def collection(self) -> Result[MongoCollection, Err]:
        if not hasattr(self, "_collection"):
            res = self.init_store()
            if res.is_err():
                return res

        return Ok(self._collection)

    @property
    def permissions(self) -> Result[MongoCollection, Err]:
        if not hasattr(self, "_permissions"):
            res = self.init_store()
            if res.is_err():
                return res

        return Ok(self._permissions)

    def set(self, *args, **kwargs):
        return self._set(*args, **kwargs)

    def _set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftObject,
        add_permissions: Optional[List[ActionObjectPermission]] = None,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        # TODO: Refactor this function since now it's doing both set and
        # update at the same time
        write_permission = ActionObjectWRITE(uid=obj.id, credentials=credentials)
        can_write = self.has_permission(write_permission)

        store_query_key: QueryKey = self.settings.store_key.with_obj(obj)
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        store_key_exists = (
            collection.find_one(store_query_key.as_dict_mongo) is not None
        )
        if (not store_key_exists) and (not self.item_keys_exist(obj, collection)):
            # attempt to claim ownership for writing
            ownership_result = self.take_ownership(uid=obj.id, credentials=credentials)
            can_write: bool = ownership_result.is_ok()
        elif not ignore_duplicates:
            unique_query_keys: QueryKeys = self.settings.unique_keys.with_obj(obj)
            keys = ", ".join(f"`{key.key}`" for key in unique_query_keys.all)
            return Err(
                f"Duplication Key Error for {obj}.\n"
                f"The fields that should be unique are {keys}."
            )
        else:
            # we are not throwing an error, because we are ignoring duplicates
            # we are also not writing though
            return Ok(obj)

        if can_write:
            storage_obj = obj.to(self.storage_type)

            collection.insert_one(storage_obj)

            # adding permissions
            read_permission = ActionObjectPermission(
                uid=obj.id,
                credentials=credentials,
                permission=ActionPermission.READ,
            )
            self.add_permission(read_permission)

            if add_permissions is not None:
                self.add_permissions(add_permissions)

            return Ok(obj)
        else:
            return Err(f"No permission to write object with id {obj.id}")

    def item_keys_exist(self, obj, collection):
        qks: QueryKeys = self.settings.unique_keys.with_obj(obj)
        query = {"$or": [{k: v} for k, v in qks.as_dict_mongo.items()]}
        res = collection.find_one(query)
        return res is not None

    def _update(
        self,
        credentials: SyftVerifyKey,
        qk: QueryKey,
        obj: SyftObject,
        has_permission: bool = False,
    ) -> Result[SyftObject, str]:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        # TODO: optimize the update. The ID should not be overwritten,
        # but the qk doesn't necessarily have to include the `id` field either.

        prev_obj_status = self._get_all_from_store(credentials, QueryKeys(qks=[qk]))
        if prev_obj_status.is_err():
            return Err(f"No object found with query key: {qk}")

        prev_obj = prev_obj_status.ok()
        if len(prev_obj) == 0:
            return Err(f"Missing values for query key: {qk}")

        prev_obj = prev_obj[0]
        if has_permission or self.has_permission(
            ActionObjectWRITE(uid=prev_obj.id, credentials=credentials)
        ):
            # we don't want to overwrite Mongo's "id_" or Syft's "id" on update
            obj_id = obj["id"]

            # Set ID to the updated object value
            obj.id = prev_obj["id"]

            # Create the Mongo object
            storage_obj = obj.to(self.storage_type)

            # revert the ID
            obj.id = obj_id

            try:
                collection.update_one(
                    filter=qk.as_dict_mongo, update={"$set": storage_obj}
                )
            except Exception as e:
                return Err(f"Failed to update obj: {obj} with qk: {qk}. Error: {e}")

            return Ok(obj)
        else:
            return Err(f"Failed to update obj {obj}, you have no permission")

    def _find_index_or_search_keys(
        self,
        credentials: SyftVerifyKey,
        index_qks: QueryKeys,
        search_qks: QueryKeys,
        order_by: Optional[PartitionKey] = None,
    ) -> Result[List[SyftObject], str]:
        # TODO: pass index as hint to find method
        qks = QueryKeys(qks=(index_qks.all + search_qks.all))
        return self._get_all_from_store(
            credentials=credentials, qks=qks, order_by=order_by
        )

    @property
    def data(self):
        values: List = self._all(credentials=None, has_permission=True).ok()
        return {v.id: v for v in values}

    def _get_all_from_store(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKeys,
        order_by: Optional[PartitionKey] = None,
        has_permission: Optional[bool] = False,
    ) -> Result[List[SyftObject], str]:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        if order_by is not None:
            storage_objs = collection.find(filter=qks.as_dict_mongo).sort(order_by.key)
        else:
            _default_key = "_id"
            storage_objs = collection.find(filter=qks.as_dict_mongo).sort(_default_key)
        syft_objs = []
        for storage_obj in storage_objs:
            obj = self.storage_type(storage_obj)
            transform_context = TransformContext(output={}, obj=obj)
            syft_objs.append(obj.to(self.settings.object_type, transform_context))

        # TODO: maybe do this in loop before this
        res = []
        for s in syft_objs:
            if has_permission or self.has_permission(
                ActionObjectREAD(uid=s.id, credentials=credentials)
            ):
                res.append(s)
        return Ok(res)

    def _delete(
        self, credentials: SyftVerifyKey, qk: QueryKey, has_permission: bool = False
    ) -> Result[SyftSuccess, Err]:
        if not (
            has_permission
            or self.has_permission(
                ActionObjectWRITE(uid=qk.value, credentials=credentials)
            )
        ):
            return Err(f"You don't have permission to delete object with qk: {qk}")

        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        collection_permissions_status = self.permissions
        if collection_permissions_status.is_err():
            return collection_permissions_status
        collection_permissions: MongoCollection = collection_permissions_status.ok()

        qks = QueryKeys(qks=qk)
        # delete the object
        result = collection.delete_one(filter=qks.as_dict_mongo)
        # delete the object's permission
        result_permission = collection_permissions.delete_one(filter=qks.as_dict_mongo)
        if result.deleted_count == 1 and result_permission.deleted_count == 1:
            return Ok(SyftSuccess(message="Object and its permission are deleted"))
        elif result.deleted_count == 0:
            return Err(f"Failed to delete object with qk: {qk}")
        else:
            return Err(
                f"Object with qk: {qk} was deleted, but failed to delete its corresponding permission"
            )

    def has_permission(self, permission: ActionObjectPermission) -> bool:
        """Check if the permission is inside the permission collection"""
        collection_permissions_status = self.permissions
        if collection_permissions_status.is_err():
            return False
        collection_permissions: MongoCollection = collection_permissions_status.ok()

        permissions: Optional[Dict] = collection_permissions.find_one(
            {"_id": permission.uid}
        )

        if permissions is None:
            return False

        # TODO: fix for other admins
        if self.root_verify_key.verify == permission.credentials.verify:
            return True

        if permission.permission_string in permissions["permissions"]:
            return True

        # check ALL_READ permission
        if (
            permission.permission == ActionPermission.READ
            and ActionObjectPermission(
                permission.uid, ActionPermission.ALL_READ
            ).permission_string
            in permissions["permissions"]
        ):
            return True

        return False

    def add_permission(self, permission: ActionObjectPermission) -> Result[None, Err]:
        collection_permissions_status = self.permissions
        if collection_permissions_status.is_err():
            return collection_permissions_status
        collection_permissions: MongoCollection = collection_permissions_status.ok()

        # find the permissions for the given permission.uid
        # e.g. permissions = {"_id": "7b88fdef6bff42a8991d294c3d66f757",
        #                      "permissions": set(["permission_str_1", "permission_str_2"]}}
        permissions: Optional[Dict] = collection_permissions.find_one(
            {"_id": permission.uid}
        )
        if permissions is None:
            # Permission doesn't exist, add a new one
            collection_permissions.insert_one(
                {
                    "_id": permission.uid,
                    "permissions": {permission.permission_string},
                }
            )
        else:
            # update the permissions with the new permission string
            permission_strings: Set = permissions["permissions"]
            permission_strings.add(permission.permission_string)
            collection_permissions.update_one(
                {"_id": permission.uid}, {"$set": {"permissions": permission_strings}}
            )

    def add_permissions(self, permissions: List[ActionObjectPermission]) -> None:
        for permission in permissions:
            self.add_permission(permission)

    def remove_permission(
        self, permission: ActionObjectPermission
    ) -> Result[None, Err]:
        collection_permissions_status = self.permissions
        if collection_permissions_status.is_err():
            return collection_permissions_status
        collection_permissions: MongoCollection = collection_permissions_status.ok()
        permissions: Optional[Dict] = collection_permissions.find_one(
            {"_id": permission.uid}
        )
        if permissions is None:
            return Err(f"permission with UID {permission.uid} not found!")
        permissions_strings: Set = permissions["permissions"]
        if permission.permission_string in permissions_strings:
            permissions_strings.remove(permission.permission_string)
            if len(permissions_strings) > 0:
                collection_permissions.update_one(
                    {"_id": permission.uid},
                    {"$set": {"permissions": permissions_strings}},
                )
            else:
                collection_permissions.delete_one({"_id": permission.uid})
        else:
            return Err(f"the permission {permission.permission_string} does not exist!")

    def take_ownership(
        self, uid: UID, credentials: SyftVerifyKey
    ) -> Result[SyftSuccess, str]:
        collection_permissions_status = self.permissions
        if collection_permissions_status.is_err():
            return collection_permissions_status
        collection_permissions: MongoCollection = collection_permissions_status.ok()

        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        data: List[UID] = collection.find_one({"_id": uid})
        permissions: List[UID] = collection_permissions.find_one({"_id": uid})

        # first person using this UID can claim ownership
        if permissions is None and data is None:
            self.add_permissions(
                [
                    ActionObjectOWNER(uid=uid, credentials=credentials),
                    ActionObjectWRITE(uid=uid, credentials=credentials),
                    ActionObjectREAD(uid=uid, credentials=credentials),
                    ActionObjectEXECUTE(uid=uid, credentials=credentials),
                ]
            )
            return Ok(SyftSuccess(message=f"Ownership of ID: {uid} taken."))

        return Err(f"UID: {uid} already owned.")

    def _all(
        self,
        credentials: SyftVerifyKey,
        order_by: Optional[PartitionKey] = None,
        has_permission: Optional[bool] = False,
    ):
        qks = QueryKeys(qks=())
        return self._get_all_from_store(
            credentials=credentials,
            qks=qks,
            order_by=order_by,
            has_permission=has_permission,
        )

    def __len__(self):
        collection_status = self.collection
        if collection_status.is_err():
            return 0
        collection: MongoCollection = collection_status.ok()
        return collection.count_documents(filter={})

    def _migrate_data(
        self, to_klass: SyftObject, context: AuthedServiceContext, has_permission: bool
    ) -> Result[bool, str]:
        credentials = context.credentials
        has_permission = (credentials == self.root_verify_key) or has_permission
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        if has_permission:
            storage_objs = collection.find({})
            for storage_obj in storage_objs:
                obj = self.storage_type(storage_obj)
                transform_context = TransformContext(output={}, obj=obj)
                value = obj.to(self.settings.object_type, transform_context)
                key = obj.get("_id")
                try:
                    migrated_value = value.migrate_to(to_klass.__version__, context)
                except Exception:
                    return Err(f"Failed to migrate data to {to_klass} for qk: {key}")
                qk = self.settings.store_key.with_obj(key)
                result = self._update(
                    credentials,
                    qk=qk,
                    obj=migrated_value,
                    has_permission=has_permission,
                )

                if result.is_err():
                    return result.err()

            return Ok(True)

        return Err("You don't have permissions to migrate data.")


@serializable()
class MongoDocumentStore(DocumentStore):
    """Mongo Document Store

    Parameters:
        `store_config`: MongoStoreConfig
            Mongo specific configuration, including connection configuration, database name, or client class type.
    """

    partition_type = MongoStorePartition


@serializable(attrs=["index_name", "settings", "store_config"])
class MongoBackingStore(KeyValueBackingStore):
    """
    Core logic for the MongoDB key-value store

    Parameters:
        `index_name`: str
            Index name (can be either 'data' or 'permissions')
        `settings`: PartitionSettings
            Syft specific settings
        `store_config`: StoreConfig
            Connection Configuration
         `ddtype`: Type
            Optional and should be None
            Used to make a consistent interface with SQLiteBackingStore
    """

    def __init__(
        self,
        index_name: str,
        settings: PartitionSettings,
        store_config: StoreConfig,
        ddtype: Optional[type] = None,
    ) -> None:
        self.index_name = index_name
        self.settings = settings
        self.store_config = store_config
        self.client: MongoClient
        self.ddtype = ddtype
        self.init_client()

    def init_client(self) -> Union[None, Err]:
        self.client = MongoClient(config=self.store_config.client_config)

        collection_status = self.client.with_collection(
            collection_settings=self.settings,
            store_config=self.store_config,
            collection_name=f"{self.settings.name}_{self.index_name}",
        )
        if collection_status.is_err():
            return collection_status
        self._collection: MongoCollection = collection_status.ok()

    @property
    def collection(self) -> Result[MongoCollection, Err]:
        if not hasattr(self, "_collection"):
            res = self.init_client()
            if res.is_err():
                return res

        return Ok(self._collection)

    def _exist(self, key: UID) -> bool:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        result: Optional[Dict] = collection.find_one({"_id": key})
        if result is not None:
            return True

        return False

    def _set(self, key: UID, value: Any) -> None:
        if self._exist(key):
            self._update(key, value)
        else:
            collection_status = self.collection
            if collection_status.is_err():
                return collection_status
            collection: MongoCollection = collection_status.ok()
            try:
                bson_data = {
                    "_id": key,
                    f"{key}": _serialize(value, to_bytes=True),
                    "_repr_debug_": _repr_debug_(value),
                }
                collection.insert_one(bson_data)
            except Exception as e:
                raise ValueError(f"Cannot insert data. Error message: {e}")

    def _update(self, key: UID, value: Any) -> None:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()
        try:
            collection.update_one(
                {"_id": key},
                {
                    "$set": {
                        f"{key}": _serialize(value, to_bytes=True),
                        "_repr_debug_": _repr_debug_(value),
                    }
                },
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to update obj: {key} with value: {value}. Error: {e}"
            )

    def __setitem__(self, key: Any, value: Any) -> None:
        self._set(key, value)

    def _get(self, key: UID) -> Any:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()

        result: Optional[Dict] = collection.find_one({"_id": key})
        if result is not None:
            return _deserialize(result[f"{key}"], from_bytes=True)
        else:
            # raise KeyError(f"{key} does not exist")
            # return an empty set which is the same with SQLiteBackingStore
            return set()

    def __getitem__(self, key: Any) -> Self:
        try:
            return self._get(key)
        except KeyError as e:
            raise e

    def _len(self) -> int:
        collection_status = self.collection
        if collection_status.is_err():
            return 0
        collection: MongoCollection = collection_status.ok()
        return collection.count_documents(filter={})

    def __len__(self) -> int:
        return self._len()

    def _delete(self, key: UID) -> None:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()
        result = collection.delete_one({"_id": key})
        if result.deleted_count != 1:
            raise KeyError(f"{key} does not exist")

    def __delitem__(self, key: str):
        self._delete(key)

    def _delete_all(self) -> None:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()
        collection.delete_many({})

    def clear(self) -> Self:
        self._delete_all()

    def _get_all(self) -> Any:
        collection_status = self.collection
        if collection_status.is_err():
            return collection_status
        collection: MongoCollection = collection_status.ok()
        result = collection.find()
        keys, values = [], []
        for row in result:
            keys.append(row["_id"])
            values.append(_deserialize(row[f"{row['_id']}"], from_bytes=True))
        return dict(zip(keys, values))

    def keys(self) -> Any:
        return self._get_all().keys()

    def values(self) -> Any:
        return self._get_all().values()

    def items(self) -> Any:
        return self._get_all().items()

    def pop(self, key: Any) -> Self:
        value = self._get(key)
        self._delete(key)
        return value

    def __contains__(self, key: Any) -> bool:
        return self._exist(key)

    def __iter__(self) -> Any:
        return iter(self.keys())

    def __repr__(self) -> str:
        return repr(self._get_all())

    def copy(self) -> Self:
        # 🟡 TODO
        raise NotImplementedError

    def update(self, *args: Any, **kwargs: Any) -> Self:
        """
        Inserts the specified items to the dictionary.
        """
        # 🟡 TODO
        raise NotImplementedError

    def __del__(self):
        """
        Close the mongo client connection:
            - Cleanup client resources and disconnect from MongoDB
            - End all server sessions created by this client
            - Close all sockets in the connection pools and stop the monitor threads
        """
        self.client.close()


@serializable()
class MongoStoreConfig(StoreConfig):
    __canonical_name__ = "MongoStoreConfig"
    """Mongo Store configuration

    Parameters:
        `client_config`: MongoStoreClientConfig
            Mongo connection details: hostname, port, user, password etc.
        `store_type`: Type[DocumentStore]
            The type of the DocumentStore. Default: MongoDocumentStore
        `db_name`: str
            Database name
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
                * FileLockingConfig: file based locking, ideal for same-device different-processes/threads stores.
                * RedisLockingConfig: Redis-based locking, ideal for multi-device stores.
            Defaults to NoLockingConfig.
    """

    client_config: MongoStoreClientConfig
    store_type: Type[DocumentStore] = MongoDocumentStore
    db_name: str = "app"
    backing_store = MongoBackingStore
    # TODO: should use a distributed lock, with RedisLockingConfig
    locking_config: LockingConfig = NoLockingConfig()



# stdlib
import importlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
import numpy as np
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...types.datetime import DateTime
from ...types.syft_object import SyftObject
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..blob_storage.service import BlobStorageService
from ..code.user_code import UserCode
from ..code.user_code import execute_byte_code
from ..context import AuthedServiceContext
from ..policy.policy import retrieve_from_db
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import UserLibConfigRegistry
from ..service import service_method
from ..user.user_roles import GUEST_ROLE_LEVEL
from .action_object import Action
from .action_object import ActionObject
from .action_object import ActionObjectPointer
from .action_object import ActionType
from .action_object import AnyActionObject
from .action_object import TwinMode
from .action_permissions import ActionObjectPermission
from .action_permissions import ActionObjectREAD
from .action_permissions import ActionPermission
from .action_store import ActionStore
from .action_types import action_type_for_type
from .numpy import NumpyArrayObject
from .pandas import PandasDataFrameObject  # noqa: F401
from .pandas import PandasSeriesObject  # noqa: F401


@serializable()
class ActionService(AbstractService):
    def __init__(self, store: ActionStore) -> None:
        self.store = store

    @service_method(path="action.np_array", name="np_array")
    def np_array(self, context: AuthedServiceContext, data: Any) -> Any:
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        np_obj = NumpyArrayObject(
            dtype=data.dtype,
            shape=data.shape,
            syft_action_data_cache=data,
            syft_node_location=context.node.id,
            syft_client_verify_key=context.credentials,
        )
        blob_store_result = np_obj._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return blob_store_result

        np_pointer = self.set(context, np_obj)
        return np_pointer

    @service_method(
        path="action.set",
        name="set",
        roles=GUEST_ROLE_LEVEL,
    )
    def set(
        self,
        context: AuthedServiceContext,
        action_object: Union[ActionObject, TwinObject],
    ) -> Result[ActionObject, str]:
        """Save an object to the action store"""
        # 🟡 TODO 9: Create some kind of type checking / protocol for SyftSerializable

        if isinstance(action_object, ActionObject):
            action_object.syft_created_at = DateTime.now()
        else:
            action_object.private_obj.syft_created_at = DateTime.now()
            action_object.mock_obj.syft_created_at = DateTime.now()

        has_result_read_permission = context.extra_kwargs.get(
            "has_result_read_permission", False
        )

        result = self.store.set(
            uid=action_object.id,
            credentials=context.credentials,
            syft_object=action_object,
            has_result_read_permission=has_result_read_permission,
        )
        if result.is_ok():
            if isinstance(action_object, TwinObject):
                action_object = action_object.mock
            action_object.syft_point_to(context.node.id)
            return Ok(action_object)
        return result.err()

    @service_method(
        path="action.is_resolved", name="is_resolved", roles=GUEST_ROLE_LEVEL
    )
    def is_resolved(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> Result[Ok[bool], Err[str]]:
        """Get an object from the action store"""
        # relative
        from .action_data_empty import ActionDataLink

        result = self._get(context, uid)
        if result.is_ok():
            obj = result.ok()
            if isinstance(obj.syft_action_data, ActionDataLink):
                result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id
                )

                # Checking in case any error occurred
                if result.is_err():
                    return result

                return Ok(result.syft_resolved)

            # If it's a leaf but not resolved yet, return false
            elif not obj.syft_resolved:
                return Ok(False)

            # If it's not an action data link or non resolved (empty). It's resolved
            return Ok(True)

        # If it's not in the store or permission error, return the error
        return result

    @service_method(
        path="action.resolve_links", name="resolve_links", roles=GUEST_ROLE_LEVEL
    )
    def resolve_links(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
    ) -> Result[Ok[ActionObject], Err[str]]:
        """Get an object from the action store"""
        # relative
        from .action_data_empty import ActionDataLink

        result = self.store.get(uid=uid, credentials=context.credentials)
        # If user has permission to get the object / object exists
        if result.is_ok():
            obj = result.ok()

            # If it's not a leaf
            if isinstance(obj.syft_action_data, ActionDataLink):
                nested_result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id, twin_mode
                )
                return nested_result

            # If it's a leaf
            return result

        return result

    @service_method(path="action.get", name="get", roles=GUEST_ROLE_LEVEL)
    def get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        resolve_nested: bool = True,
    ) -> Result[Ok[ActionObject], Err[str]]:
        """Get an object from the action store"""
        return self._get(context, uid, twin_mode, resolve_nested=resolve_nested)

    def _get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        has_permission=False,
        resolve_nested: bool = True,
    ) -> Result[ActionObject, str]:
        """Get an object from the action store"""
        # stdlib

        # relative
        from .action_data_empty import ActionDataLink

        result = self.store.get(
            uid=uid, credentials=context.credentials, has_permission=has_permission
        )
        if result.is_ok():
            obj: Union[TwinObject, ActionObject] = result.ok()
            obj._set_obj_location_(
                context.node.id,
                context.credentials,
            )
            # Resolve graph links
            if (
                not isinstance(obj, TwinObject)
                and resolve_nested
                and isinstance(obj.syft_action_data, ActionDataLink)
            ):
                if not self.is_resolved(
                    context, obj.syft_action_data.action_object_id.id
                ).ok():
                    return SyftError(message="This object is not resolved yet.")
                result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id, twin_mode
                )
                return result
            if isinstance(obj, TwinObject):
                if twin_mode == TwinMode.PRIVATE:
                    obj = obj.private
                    obj.syft_point_to(context.node.id)
                elif twin_mode == TwinMode.MOCK:
                    obj = obj.mock
                    obj.syft_point_to(context.node.id)
                else:
                    obj.mock.syft_point_to(context.node.id)
                    obj.private.syft_point_to(context.node.id)
            return Ok(obj)
        else:
            return result

    @service_method(
        path="action.get_pointer", name="get_pointer", roles=GUEST_ROLE_LEVEL
    )
    def get_pointer(
        self, context: AuthedServiceContext, uid: UID
    ) -> Result[ActionObjectPointer, str]:
        """Get a pointer from the action store"""
        result = self.store.get_pointer(
            uid=uid, credentials=context.credentials, node_uid=context.node.id
        )
        if result.is_ok():
            obj = result.ok()
            obj._set_obj_location_(
                context.node.id,
                context.credentials,
            )
            return Ok(obj)
        return Err(result.err())

    @service_method(path="action.get_mock", name="get_mock", roles=GUEST_ROLE_LEVEL)
    def get_mock(
        self, context: AuthedServiceContext, uid: UID
    ) -> Result[SyftError, SyftObject]:
        """Get a pointer from the action store"""
        result = self.store.get_mock(uid=uid)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    # not a public service endpoint
    def _user_code_execute(
        self,
        context: AuthedServiceContext,
        code_item: UserCode,
        kwargs: Dict[str, Any],
        result_id: Optional[UID] = None,
    ) -> Result[ActionObjectPointer, Err]:
        if not context.has_execute_permissions:
            input_policy = code_item.input_policy
            filtered_kwargs = input_policy.filter_kwargs(
                kwargs=kwargs, context=context, code_item_id=code_item.id
            )
            if isinstance(filtered_kwargs, SyftError) or filtered_kwargs.is_err():
                return filtered_kwargs
            filtered_kwargs = filtered_kwargs.ok()
        else:
            filtered_kwargs = retrieve_from_db(code_item.id, kwargs, context).ok()
        # update input policy to track any input state
        # code_item.input_policy = input_policy

        if not context.has_execute_permissions:
            expected_input_kwargs = set()
            for _inp_kwarg in code_item.input_policy.inputs.values():
                keys = _inp_kwarg.keys()
                for k in keys:
                    if k not in kwargs:
                        return Err(
                            f"{code_item.service_func_name}() missing required keyword argument: '{k}'"
                        )
                expected_input_kwargs.update(keys)

            permitted_input_kwargs = list(filtered_kwargs.keys())
            not_approved_kwargs = set(expected_input_kwargs) - set(
                permitted_input_kwargs
            )
            if len(not_approved_kwargs) > 0:
                return Err(
                    f"Input arguments: {not_approved_kwargs} to the function are not approved yet."
                )

        has_twin_inputs = False

        real_kwargs = {}
        for key, kwarg_value in filtered_kwargs.items():
            if isinstance(kwarg_value, TwinObject):
                has_twin_inputs = True
            real_kwargs[key] = kwarg_value

        result_id = UID() if result_id is None else result_id

        try:
            if not has_twin_inputs:
                # no twins
                filtered_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.NONE
                )
                exec_result = execute_byte_code(code_item, filtered_kwargs, context)
                if isinstance(exec_result.result, ActionObject):
                    result_action_object = ActionObject.link(
                        result_id=result_id, pointer_id=exec_result.result.id
                    )
                else:
                    result_action_object = wrap_result(result_id, exec_result.result)
            else:
                # twins
                private_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.PRIVATE
                )
                private_exec_result = execute_byte_code(
                    code_item, private_kwargs, context
                )
                result_action_object_private = wrap_result(
                    result_id, private_exec_result.result
                )

                mock_kwargs = filter_twin_kwargs(real_kwargs, twin_mode=TwinMode.MOCK)
                # relative
                from .action_data_empty import ActionDataEmpty

                if any(isinstance(v, ActionDataEmpty) for v in mock_kwargs.values()):
                    mock_exec_result_obj = ActionDataEmpty()
                else:
                    mock_exec_result = execute_byte_code(
                        code_item, mock_kwargs, context
                    )
                    mock_exec_result_obj = mock_exec_result.result

                result_action_object_mock = wrap_result(result_id, mock_exec_result_obj)

                result_action_object = TwinObject(
                    id=result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
        except Exception as e:
            # import traceback
            # return Err(f"_user_code_execute failed. {e} {traceback.format_exc()}")
            return Err(f"_user_code_execute failed. {e}")
        return Ok(result_action_object)

    def set_result_to_store(self, result_action_object, context, output_policy=None):
        result_id = result_action_object.id
        # result_blob_id = result_action_object.syft_blob_storage_entry_id

        if output_policy is not None:
            output_readers = (
                output_policy.output_readers
                if not context.has_execute_permissions
                else []
            )
        else:
            output_readers = []

        read_permission = ActionPermission.READ

        result_action_object._set_obj_location_(
            context.node.id,
            context.credentials,
        )
        blob_store_result = result_action_object._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return blob_store_result

        # IMPORTANT: DO THIS ONLY AFTER ._save_to_blob_storage
        if isinstance(result_action_object, TwinObject):
            result_blob_id = result_action_object.private.syft_blob_storage_entry_id
        else:
            result_blob_id = result_action_object.syft_blob_storage_entry_id

        # pass permission information to the action store as extra kwargs
        context.extra_kwargs = {"has_result_read_permission": True}

        set_result = self.set(context, result_action_object)

        if set_result.is_err():
            return set_result

        blob_storage_service: BlobStorageService = context.node.get_service(
            BlobStorageService
        )

        def store_permission(x):
            return ActionObjectPermission(result_id, read_permission, x)

        def blob_permission(x):
            return ActionObjectPermission(result_blob_id, read_permission, x)

        if len(output_readers) > 0:
            store_permissions = [store_permission(x) for x in output_readers]
            self.store.add_permissions(store_permissions)

            blob_permissions = [blob_permission(x) for x in output_readers]
            blob_storage_service.stash.add_permissions(blob_permissions)

        return set_result

    def execute_plan(
        self, plan, context: AuthedServiceContext, plan_kwargs: Dict[str, ActionObject]
    ):
        id2inpkey = {v.id: k for k, v in plan.inputs.items()}

        for plan_action in plan.actions:
            if (
                hasattr(plan_action.remote_self, "id")
                and plan_action.remote_self.id in id2inpkey
            ):
                plan_action.remote_self = plan_kwargs[
                    id2inpkey[plan_action.remote_self.id]
                ]
            for i, arg in enumerate(plan_action.args):
                if arg in id2inpkey:
                    plan_action.args[i] = plan_kwargs[id2inpkey[arg]]

            for k, arg in enumerate(plan_action.kwargs):
                if arg in id2inpkey:
                    plan_action.kwargs[k] = plan_kwargs[id2inpkey[arg]]

        for plan_action in plan.actions:
            action_res = self.execute(context, plan_action)
            if isinstance(action_res, SyftError):
                return action_res
        result_id = plan.outputs[0].id
        return self._get(context, result_id, TwinMode.MOCK, has_permission=True)

    def call_function(self, context: AuthedServiceContext, action: Action):
        # run function/class init
        _user_lib_config_registry = UserLibConfigRegistry.from_user(context.credentials)
        absolute_path = f"{action.path}.{action.op}"
        if absolute_path in _user_lib_config_registry:
            # TODO: implement properly
            # Now we are assuming its a function/class
            return execute_callable(self, context, action)
        else:
            return Err(
                f"Failed executing {action}. You have no permission for {absolute_path}"
            )

    def set_attribute(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: Union[ActionObject, TwinObject],
    ):
        args, _ = resolve_action_args(action, context, self)
        if args.is_err():
            return Err(
                f"Failed executing action {action}, could not resolve args: {args.err()}"
            )
        else:
            args = args.ok()
        if not isinstance(args[0], ActionObject):
            return Err(
                f"Failed executing action {action} setattribute requires a non-twin string as first argument"
            )
        name = args[0].syft_action_data
        # dont do the whole filtering dance with the name
        args = [args[1]]

        if isinstance(resolved_self, TwinObject):
            # todo, create copy?
            private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE)
            private_val = private_args[0]
            setattr(resolved_self.private.syft_action_data, name, private_val)
            # todo: what do we use as data for the mock here?
            # depending on permisisons?
            public_args = filter_twin_args(args, twin_mode=TwinMode.MOCK)
            public_val = public_args[0]
            setattr(resolved_self.mock.syft_action_data, name, public_val)
            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=ActionObject.from_obj(
                        resolved_self.private.syft_action_data
                    ),
                    private_obj_id=action.result_id,
                    mock_obj=ActionObject.from_obj(resolved_self.mock.syft_action_data),
                    mock_obj_id=action.result_id,
                )
            )
        else:
            # TODO: Implement for twinobject args
            args = filter_twin_args(args, twin_mode=TwinMode.NONE)
            val = args[0]
            setattr(resolved_self.syft_action_data, name, val)
            return Ok(
                ActionObject.from_obj(resolved_self.syft_action_data),
            )
            # todo: permissions
            # setattr(resolved_self.syft_action_data, name, val)
            # val = resolved_self.syft_action_data
            # result_action_object = Ok(wrap_result(action.result_id, val))

    def get_attribute(
        self, action: Action, resolved_self: Union[ActionObject, TwinObject]
    ):
        if isinstance(resolved_self, TwinObject):
            private_result = getattr(resolved_self.private.syft_action_data, action.op)
            mock_result = getattr(resolved_self.mock.syft_action_data, action.op)
            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=ActionObject.from_obj(private_result),
                    private_obj_id=action.result_id,
                    mock_obj=ActionObject.from_obj(mock_result),
                    mock_obj_id=action.result_id,
                )
            )
        else:
            val = getattr(resolved_self.syft_action_data, action.op)
            return Ok(wrap_result(action.result_id, val))

    def call_method(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: Union[ActionObject, TwinObject],
    ):
        if isinstance(resolved_self, TwinObject):
            # method
            private_result = execute_object(
                self,
                context,
                resolved_self.private,
                action,
                twin_mode=TwinMode.PRIVATE,
            )
            if private_result.is_err():
                return Err(
                    f"Failed executing action {action}, result is an error: {private_result.err()}"
                )
            mock_result = execute_object(
                self, context, resolved_self.mock, action, twin_mode=TwinMode.MOCK
            )
            if mock_result.is_err():
                return Err(
                    f"Failed executing action {action}, result is an error: {mock_result.err()}"
                )

            private_result = private_result.ok()
            mock_result = mock_result.ok()

            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=private_result,
                    private_obj_id=action.result_id,
                    mock_obj=mock_result,
                    mock_obj_id=action.result_id,
                )
            )
        else:
            return execute_object(self, context, resolved_self, action)

    @service_method(path="action.execute", name="execute", roles=GUEST_ROLE_LEVEL)
    def execute(
        self, context: AuthedServiceContext, action: Action
    ) -> Result[ActionObject, Err]:
        """Execute an operation on objects in the action store"""
        # relative
        from .plan import Plan

        if action.action_type == ActionType.CREATEOBJECT:
            result_action_object = Ok(action.create_object)
            # print(action.create_object, "already in blob storage")
        elif action.action_type == ActionType.SYFTFUNCTION:
            usercode_service = context.node.get_service("usercodeservice")
            kwarg_ids = {}
            for k, v in action.kwargs.items():
                # transform lineage ids into ids
                kwarg_ids[k] = v.id
            result_action_object: Result[ActionObject, Err] = usercode_service._call(
                context, action.user_code_id, action.result_id, **kwarg_ids
            )
            return result_action_object
        elif action.action_type == ActionType.FUNCTION:
            result_action_object = self.call_function(context, action)
        else:
            resolved_self = self._get(
                context=context,
                uid=action.remote_self,
                twin_mode=TwinMode.NONE,
                has_permission=True,
            )
            if resolved_self.is_err():
                return Err(
                    f"Failed executing action {action}, could not resolve self: {resolved_self.err()}"
                )
            resolved_self = resolved_self.ok()
            if action.op == "__call__" and resolved_self.syft_action_data_type == Plan:
                result_action_object = self.execute_plan(
                    plan=resolved_self.syft_action_data,
                    context=context,
                    plan_kwargs=action.kwargs,
                )
                return result_action_object
            elif action.action_type == ActionType.SETATTRIBUTE:
                result_action_object = self.set_attribute(
                    context, action, resolved_self
                )
            elif action.action_type == ActionType.GETATTRIBUTE:
                result_action_object = self.get_attribute(action, resolved_self)
            elif action.action_type == ActionType.METHOD:
                result_action_object = self.call_method(context, action, resolved_self)
            else:
                return Err("Unknown action")

        if result_action_object.is_err():
            return Err(
                f"Failed executing action {action}, result is an error: {result_action_object.err()}"
            )
        else:
            result_action_object = result_action_object.ok()

        # check if we have read permissions on the result
        has_result_read_permission = self.has_read_permission_for_action_result(
            context, action
        )

        result_action_object._set_obj_location_(
            context.node.id,
            context.credentials,
        )

        blob_store_result = result_action_object._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return blob_store_result

        # pass permission information to the action store as extra kwargs
        context.extra_kwargs = {
            "has_result_read_permission": has_result_read_permission
        }

        set_result = self.set(context, result_action_object)
        if set_result.is_err():
            return Err(
                f"Failed executing action {action}, set result is an error: {set_result.err()}"
            )

        return set_result

    def has_read_permission_for_action_result(
        self, context: AuthedServiceContext, action: Action
    ) -> bool:
        action_obj_ids = (
            action.args + list(action.kwargs.values()) + [action.remote_self]
        )
        permissions = [
            ActionObjectREAD(uid=_id, credentials=context.credentials)
            for _id in action_obj_ids
        ]
        return self.store.has_permissions(permissions)

    @service_method(path="action.exists", name="exists", roles=GUEST_ROLE_LEVEL)
    def exists(
        self, context: AuthedServiceContext, obj_id: UID
    ) -> Result[SyftSuccess, SyftError]:
        """Checks if the given object id exists in the Action Store"""
        if self.store.exists(obj_id):
            return SyftSuccess(message=f"Object: {obj_id} exists")
        else:
            return SyftError(message=f"Object: {obj_id} does not exist")


def resolve_action_args(
    action: Action, context: AuthedServiceContext, service: ActionService
):
    has_twin_inputs = False
    args = []
    for arg_id in action.args:
        arg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        )
        if arg_value.is_err():
            return arg_value, False
        if isinstance(arg_value.ok(), TwinObject):
            has_twin_inputs = True
        args.append(arg_value.ok())
    return Ok(args), has_twin_inputs


def resolve_action_kwargs(
    action: Action, context: AuthedServiceContext, service: ActionService
):
    has_twin_inputs = False
    kwargs = {}
    for key, arg_id in action.kwargs.items():
        kwarg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        )
        if kwarg_value.is_err():
            return kwarg_value, False
        if isinstance(kwarg_value.ok(), TwinObject):
            has_twin_inputs = True
        kwargs[key] = kwarg_value.ok()
    return Ok(kwargs), has_twin_inputs


def execute_callable(
    service: ActionService,
    context: AuthedServiceContext,
    action: Action,
) -> Result[ActionObject, str]:
    args, has_arg_twins = resolve_action_args(action, context, service)
    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service)
    has_twin_inputs = has_arg_twins or has_kwargs_twins
    if args.is_err():
        return args
    else:
        args = args.ok()
    if kwargs.is_err():
        return kwargs
    else:
        kwargs = kwargs.ok()

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    # stdlib

    # TODO: get from CMPTree is probably safer
    def _get_target_callable(path: str, op: str):
        path_elements = path.split(".")
        res = importlib.import_module(path_elements[0])
        for p in path_elements[1:]:
            res = getattr(res, p)
        res = getattr(res, op)
        return res

    target_callable = _get_target_callable(action.path, action.op)

    result = None
    try:
        if target_callable:
            if not has_twin_inputs:
                # if twin_mode == TwinMode.NONE and not has_twin_inputs:
                twin_mode = TwinMode.NONE
                # no twins
                filtered_args = filter_twin_args(args, twin_mode=twin_mode)
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_callable(*filtered_args, **filtered_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            else:
                twin_mode = TwinMode.PRIVATE
                private_args = filter_twin_args(args, twin_mode=twin_mode)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                private_result = target_callable(*private_args, **private_kwargs)
                result_action_object_private = wrap_result(
                    action.result_id, private_result
                )

                twin_mode = TwinMode.MOCK
                mock_args = filter_twin_args(args, twin_mode=twin_mode)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                mock_result = target_callable(*mock_args, **mock_kwargs)
                result_action_object_mock = wrap_result(action.result_id, mock_result)

                result_action_object = TwinObject(
                    id=action.result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )

    except Exception as e:
        print("what is this exception", e)
        return Err(e)
    return Ok(result_action_object)


def execute_object(
    service: ActionService,
    context: AuthedServiceContext,
    resolved_self: ActionObject,
    action: Action,
    twin_mode: TwinMode = TwinMode.NONE,
) -> Result[Ok[Union[TwinObject, ActionObject]], Err[str]]:
    unboxed_resolved_self = resolved_self.syft_action_data
    _args, has_arg_twins = resolve_action_args(action, context, service)

    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service)
    if _args.is_err():
        return _args
    else:
        args = _args.ok()
    if kwargs.is_err():
        return kwargs
    else:
        kwargs = kwargs.ok()
    has_twin_inputs = has_arg_twins or has_kwargs_twins

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    target_method = getattr(unboxed_resolved_self, action.op, None)
    result = None
    try:
        if target_method:
            if twin_mode == TwinMode.NONE and not has_twin_inputs:
                # no twins
                filtered_args = filter_twin_args(args, twin_mode=twin_mode)
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*filtered_args, **filtered_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            elif twin_mode == TwinMode.NONE and has_twin_inputs:
                # self isn't a twin but one of the inputs is
                private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.PRIVATE)
                private_result = target_method(*private_args, **private_kwargs)
                result_action_object_private = wrap_result(
                    action.result_id, private_result
                )

                mock_args = filter_twin_args(args, twin_mode=TwinMode.MOCK)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.MOCK)
                mock_result = target_method(*mock_args, **mock_kwargs)
                result_action_object_mock = wrap_result(action.result_id, mock_result)

                result_action_object = TwinObject(
                    id=action.result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
            elif twin_mode == twin_mode.PRIVATE:  # type: ignore
                # twin private path
                private_args = filter_twin_args(args, twin_mode=twin_mode)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*private_args, **private_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            elif twin_mode == twin_mode.MOCK:  # type: ignore
                # twin mock path
                mock_args = filter_twin_args(args, twin_mode=twin_mode)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                target_method = getattr(unboxed_resolved_self, action.op, None)
                result = target_method(*mock_args, **mock_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            else:
                raise Exception(
                    f"Bad combination of: twin_mode: {twin_mode} and has_twin_inputs: {has_twin_inputs}"
                )
        else:
            return Err("Missing target method")

    except Exception as e:
        return Err(e)

    return Ok(result_action_object)


def wrap_result(result_id: UID, result: Any) -> ActionObject:
    # 🟡 TODO 11: Figure out how we want to store action object results
    action_type = action_type_for_type(result)
    result_action_object = action_type(id=result_id, syft_action_data_cache=result)
    return result_action_object


def filter_twin_args(args: List[Any], twin_mode: TwinMode) -> Any:
    filtered = []
    for arg in args:
        if isinstance(arg, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered.append(arg.private.syft_action_data)
            elif twin_mode == TwinMode.MOCK:
                filtered.append(arg.mock.syft_action_data)
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered.append(arg.syft_action_data)
    return filtered


def filter_twin_kwargs(kwargs: Dict, twin_mode: TwinMode) -> Any:
    filtered = {}
    for k, v in kwargs.items():
        if isinstance(v, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered[k] = v.private.syft_action_data
            elif twin_mode == TwinMode.MOCK:
                filtered[k] = v.mock.syft_action_data
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered[k] = v.syft_action_data
    return filtered


TYPE_TO_SERVICE[ActionObject] = ActionService
TYPE_TO_SERVICE[TwinObject] = ActionService
TYPE_TO_SERVICE[AnyActionObject] = ActionService

SERVICE_TO_TYPES[ActionService].update({ActionObject, TwinObject, AnyActionObject})



# stdlib
from typing import List
from typing import Tuple
from typing import Union

# relative
from ..service.dataset.dataset import Dataset
from ..service.metadata.node_metadata import NodeMetadataJSON
from ..service.network.network_service import NodePeer
from ..types.uid import UID
from .client import SyftClient
from .registry import DomainRegistry


class SearchResults:
    def __init__(self, results: List[Tuple[SyftClient, List[Dataset]]]) -> None:
        self._dataset_client = {}
        self._datasets = []
        for pairs in results:
            client = pairs[0]
            datasets = list(pairs[1])
            for dataset in datasets:
                self._dataset_client[dataset.id] = client
                self._datasets.append(dataset)

    def __getitem__(self, key: Union[int, str, UID]) -> Dataset:
        if isinstance(key, int):
            return self._datasets[key]
        else:
            for dataset in self._datasets:
                if isinstance(key, UID):
                    if dataset.id == key:
                        return dataset
                elif isinstance(key, str):
                    if dataset.name == key:
                        return dataset
                    elif str(dataset.id) == key:
                        return dataset
        raise KeyError

    def __repr__(self) -> str:
        return str(self._datasets)

    def _repr_html_(self) -> str:
        return self._datasets._repr_html_()

    def client_for(self, key: Union[Dataset, int, str, UID]) -> SyftClient:
        if isinstance(key, Dataset):
            dataset = key
        else:
            dataset = self.__getitem__(key)
        return self._dataset_client[dataset.id]


class Search:
    def __init__(self, domains: DomainRegistry):
        self.domains = domains.online_domains

    @staticmethod
    def __search_one_node(
        peer_tuple: Tuple[NodePeer, NodeMetadataJSON], name: str
    ) -> List[Dataset]:
        try:
            peer, _ = peer_tuple
            client = peer.guest_client
            results = client.api.services.dataset.search(name=name)
            return (client, results)
        except:  # noqa
            return (None, [])

    def __search(self, name: str) -> List[Tuple[SyftClient, List[Dataset]]]:
        results = [
            self.__search_one_node(peer_tuple, name) for peer_tuple in self.domains
        ]

        # filter out SyftError
        filtered = ((client, result) for client, result in results if result)
        return filtered

    def search(self, name: str) -> SearchResults:
        return SearchResults(self.__search(name))



# stdlib
from enum import Enum
from typing import List
from typing import Optional

# relative
from ...client.api import APIRegistry
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import add_credentials_for_key
from ...types.transforms import add_node_uid_for_key
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE


@serializable()
class NotificationStatus(Enum):
    UNREAD = 0
    READ = 1


@serializable()
class NotificationRequestStatus(Enum):
    NO_ACTION = 0


class NotificationExpiryStatus(Enum):
    AUTO = 0
    NEVER = 1


@serializable()
class ReplyNotification(SyftObject):
    __canonical_name__ = "ReplyNotification"
    __version__ = SYFT_OBJECT_VERSION_1

    text: str
    target_msg: UID
    id: Optional[UID]
    from_user_verify_key: Optional[SyftVerifyKey]


@serializable()
class Notification(SyftObject):
    __canonical_name__ = "Notification"
    __version__ = SYFT_OBJECT_VERSION_1

    subject: str
    node_uid: UID
    from_user_verify_key: SyftVerifyKey
    to_user_verify_key: SyftVerifyKey
    created_at: DateTime
    status: NotificationStatus = NotificationStatus.UNREAD
    linked_obj: Optional[LinkedObject]
    replies: Optional[List[ReplyNotification]] = []

    __attr_searchable__ = [
        "from_user_verify_key",
        "to_user_verify_key",
        "status",
    ]
    __repr_attrs__ = ["subject", "status", "created_at", "linked_obj"]

    def _repr_html_(self) -> str:
        return f"""
            <style>
            .syft-request {{color: {SURFACE[options.color_theme]}; line-height: 1;}}
            </style>
            <div class='syft-request'>
                <h3>Notification</h3>
                <p><strong>ID: </strong>{self.id}</p>
                <p><strong>Subject: </strong>{self.subject}</p>
                <p><strong>Status: </strong>{self.status.name}</p>
                <p><strong>Created at: </strong>{self.created_at}</p>
                <p><strong>Linked object: </strong>{self.linked_obj}</p>
                <p>
            </div>
        """

    @property
    def link(self) -> Optional[SyftObject]:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None

    def _coll_repr_(self):
        return {
            "Subject": self.subject,
            "Status": self.determine_status().name.capitalize(),
            "Created At": str(self.created_at),
            "Linked object": f"{self.linked_obj.object_type.__canonical_name__} ({self.linked_obj.object_uid})",
        }

    def mark_read(self) -> None:
        api = APIRegistry.api_for(
            self.node_uid, user_verify_key=self.syft_client_verify_key
        )
        return api.services.notifications.mark_as_read(uid=self.id)

    def mark_unread(self) -> None:
        api = APIRegistry.api_for(
            self.node_uid, user_verify_key=self.syft_client_verify_key
        )
        return api.services.notifications.mark_as_unread(uid=self.id)

    def determine_status(self) -> Enum:
        # relative
        from ..request.request import Request

        if isinstance(self.linked_obj.resolve, Request):
            return self.linked_obj.resolve.status

        return NotificationRequestStatus.NO_ACTION


@serializable()
class CreateNotification(Notification):
    __canonical_name__ = "CreateNotification"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    node_uid: Optional[UID]
    from_user_verify_key: Optional[SyftVerifyKey]
    created_at: Optional[DateTime]


def add_msg_creation_time(context: TransformContext) -> TransformContext:
    context.output["created_at"] = DateTime.now()
    return context


@transform(CreateNotification, Notification)
def createnotification_to_notification():
    return [
        generate_id,
        add_msg_creation_time,
        add_credentials_for_key("from_user_verify_key"),
        add_node_uid_for_key("node_uid"),
    ]



# stdlib
import contextlib
import os.path
from pathlib import Path
from typing import Tuple

# third party
import docker

# relative
from .config import CustomWorkerConfig


class CustomWorkerBuilder:
    TYPE_CPU = "cpu"
    TYPE_GPU = "gpu"

    DOCKERFILE_PROD_PATH = os.path.expandvars("$APPDIR/grid/")
    DOCKERFILE_DEV_PATH = "../../../../../grid/backend/"

    CUSTOM_IMAGE_PREFIX = "custom-worker"

    BUILD_MAX_WAIT = 30 * 60

    def build_image(self, config: CustomWorkerConfig) -> None:
        """
        Builds a Docker image for the custom worker based on the provided configuration.
        Args:
            config (CustomImageConfig): The configuration for building the Docker image.
        Returns:
            bool: True if the image was built successfully, raises Exception otherwise.
        """

        # remove once GPU is supported
        if config.build.gpu:
            raise Exception("GPU custom worker is not supported yet")

        type = self.TYPE_GPU if config.build.gpu else self.TYPE_CPU

        contextdir, dockerfile = self.find_worker_ctx(type)

        imgtag = config.get_signature()[:8]

        build_args = {
            "PYTHON_VERSION": config.build.python_version,
            "SYSTEM_PACKAGES": config.build.merged_system_pkgs(),
            "PIP_PACKAGES": config.build.merged_python_pkgs(),
            "CUSTOM_CMD": config.build.merged_custom_cmds(),
        }

        print(
            f"Building dockerfile={dockerfile} "
            f"in context={contextdir} "
            f"with args={build_args}"
        )

        try:
            # TODO: Push logs to mongo/seaweed?
            with contextlib.closing(docker.from_env()) as client:
                client.images.build(
                    path=str(contextdir),
                    dockerfile=dockerfile,
                    pull=True,
                    tag=f"{self.CUSTOM_IMAGE_PREFIX}-{type}:{imgtag}",
                    timeout=self.BUILD_MAX_WAIT,
                    buildargs=build_args,
                )

            return
        except docker.errors.BuildError as e:
            raise e
        except docker.errors.APIError as e:
            raise e
        except Exception as e:
            raise e

    def find_worker_ctx(self, type: str) -> Tuple[Path, str]:
        """
        Find the Worker Dockerfile and it's context path
        - PROD will be in `$APPDIR/grid/`
        - DEV will be in `packages/grid/backend`
        - In both the cases context dir does not matter (unless we're calling COPY)

        Args:
            type (str): The type of worker.
        Returns:
            Path: The path to the Dockerfile.
        """
        filename = f"worker_{type}.dockerfile"
        lookup_paths = [
            Path(self.DOCKERFILE_PROD_PATH, filename).resolve(),
            Path(__file__, self.DOCKERFILE_DEV_PATH, filename).resolve(),
        ]
        for path in lookup_paths:
            if path.exists():
                return path.parent, filename

        raise FileNotFoundError(f"Cannot find the {filename}")



# relative
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject

ParentPartitionKey = PartitionKey(key="parent", type_=str)
ChildPartitionKey = PartitionKey(key="child", type_=str)


@serializable()
class DataSubjectMemberRelationship(SyftObject):
    __canonical_name__ = "DataSubjectMemberRelationship"
    __version__ = SYFT_OBJECT_VERSION_1

    parent: str
    child: str

    __attr_searchable__ = ["parent", "child"]
    __attr_unique__ = ["parent", "child"]

    def __hash__(self) -> int:
        return hash(self.parent + self.child)

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    def __repr__(self) -> str:
        return f"<DataSubjectMembership: {self.parent} -> {self.child}>"



# stdlib
from typing import List
from typing import Optional

# third party
from result import Result

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from .code_history import CodeHistory

NamePartitionKey = PartitionKey(key="service_func_name", type_=str)
VerifyKeyPartitionKey = PartitionKey(key="user_verify_key", type_=SyftVerifyKey)


@serializable()
class CodeHistoryStash(BaseUIDStoreStash):
    object_type = CodeHistory
    settings: PartitionSettings = PartitionSettings(
        name=CodeHistory.__canonical_name__, object_type=CodeHistory
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_by_service_func_name_and_verify_key(
        self,
        credentials: SyftVerifyKey,
        service_func_name: str,
        user_verify_key: SyftVerifyKey,
    ) -> Result[List[CodeHistory], str]:
        qks = QueryKeys(
            qks=[
                NamePartitionKey.with_obj(service_func_name),
                VerifyKeyPartitionKey.with_obj(user_verify_key),
            ]
        )
        return self.query_one(credentials=credentials, qks=qks)

    def get_by_service_func_name(
        self, credentials: SyftVerifyKey, service_func_name: str
    ) -> Result[List[CodeHistory], str]:
        qks = QueryKeys(qks=[NamePartitionKey.with_obj(service_func_name)])
        return self.query_all(credentials=credentials, qks=qks)

    def get_by_verify_key(
        self, credentials: SyftVerifyKey, user_verify_key: SyftVerifyKey
    ) -> Result[Optional[CodeHistory], str]:
        if isinstance(user_verify_key, str):
            user_verify_key = SyftVerifyKey.from_string(user_verify_key)
        qks = QueryKeys(qks=[VerifyKeyPartitionKey.with_obj(user_verify_key)])
        return self.query_all(credentials=credentials, qks=qks)

    # def get_version(self, name:str, version:int) -> Optional[UserCode]:
    #     for obj in self.objs.values():
    #         if obj.name == name and obj.version == version:
    #             return obj
    #     return None



# stdlib
import contextlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

# third party
import docker
from docker.models.containers import Container

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from .utils import run_containers
from .worker_image_stash import SyftWorkerImageStash
from .worker_pool import ContainerSpawnStatus
from .worker_pool import SyftWorker
from .worker_pool import WorkerOrchestrationType
from .worker_pool import WorkerPool
from .worker_pool import WorkerStatus
from .worker_pool_stash import SyftWorkerPoolStash


@serializable()
class SyftWorkerPoolService(AbstractService):
    store: DocumentStore
    stash: SyftWorkerPoolStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftWorkerPoolStash(store=store)
        self.image_stash = SyftWorkerImageStash(store=store)

    @service_method(
        path="worker_pool.create",
        name="create",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def create_pool(
        self,
        context: AuthedServiceContext,
        name: str,
        image_uid: UID,
        number: int,
    ) -> Union[List[ContainerSpawnStatus], SyftError]:
        """Creates a pool of workers from the given SyftWorkerImage.

        - Retrieves the image for the given UID
        - Use docker to launch containers for given image
        - For each successful container instantiation create a SyftWorker object
        - Creates a SyftWorkerPool object

        Args:
            context (AuthedServiceContext): context passed to the service
            name (str): name of the pool
            image_id (UID): UID of the SyftWorkerImage against which the pool should be created
            number (int): number of SyftWorker that needs to be created in the pool
        """

        result = self.stash.get_by_name(context.credentials, pool_name=name)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        if result.ok() is not None:
            return SyftError(message=f"Worker Pool with name: {name} already exists !!")

        result = self.image_stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        )
        if result.is_err():
            return SyftError(
                message=f"Failed to retrieve Worker Image with id: {image_uid}. Error: {result.err()}"
            )

        worker_image = result.ok()

        container_statuses: List[ContainerSpawnStatus] = run_containers(
            pool_name=name,
            worker_image=worker_image,
            number=number,
            orchestration=WorkerOrchestrationType.DOCKER,
        )

        workers = [
            container_status.worker
            for container_status in container_statuses
            if container_status.worker
        ]

        worker_pool = WorkerPool(
            name=name,
            syft_worker_image_id=image_uid,
            max_count=number,
            workers=workers,
        )
        result = self.stash.set(credentials=context.credentials, obj=worker_pool)

        if result.is_err():
            return SyftError(message=f"Failed to save Worker Pool: {result.err()}")

        return container_statuses

    @service_method(
        path="worker_pool.get_all",
        name="get_all",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_all(
        self, context: AuthedServiceContext
    ) -> Union[List[WorkerPool], SyftError]:
        # TODO: During get_all, we should dynamically make a call to docker to get the status of the containers
        # and update the status of the workers in the pool.
        result = self.stash.get_all(credentials=context.credentials)
        if result.is_err():
            return SyftError(message=f"{result.err()}")

        return result.ok()

    @service_method(
        path="worker_pool.delete_worker",
        name="delete_worker",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def delete_worker(
        self,
        context: AuthedServiceContext,
        worker_pool_id: UID,
        worker_id: UID,
        force: bool = False,
    ) -> Union[SyftSuccess, SyftError]:
        worker_pool_worker = self._get_worker_pool_and_worker(
            context, worker_pool_id, worker_id
        )
        if isinstance(worker_pool_worker, SyftError):
            return worker_pool_worker

        worker_pool, worker = worker_pool_worker

        # delete the worker using docker client sdk
        with contextlib.closing(docker.from_env()) as client:
            docker_container = _get_worker_container(client, worker)
            if isinstance(docker_container, SyftError):
                return docker_container

            stopped = _stop_worker_container(worker, docker_container, force)
            if stopped is not None:
                return stopped

        # remove the worker from the pool
        worker_pool.workers.remove(worker)
        result = self.stash.update(context.credentials, obj=worker_pool)
        if result.is_err():
            return SyftError(message=f"Failed to update worker pool: {result.err()}")

        return SyftSuccess(
            message=f"Worker with id: {worker_id} deleted successfully from pool: {worker_pool.name}"
        )

    @service_method(
        path="worker_pool.get_worker",
        name="get_worker",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_worker(
        self, context: AuthedServiceContext, worker_pool_id: UID, worker_id: UID
    ) -> Union[SyftWorker, SyftError]:
        worker_pool_worker = self._get_worker_pool_and_worker(
            context, worker_pool_id, worker_id
        )
        if isinstance(worker_pool_worker, SyftError):
            return worker_pool_worker

        worker_pool, worker = worker_pool_worker

        with contextlib.closing(docker.from_env()) as client:
            container = _get_worker_container(client, worker)
            if isinstance(container, SyftError):
                return container

            container_status = container.status

        worker_status = _CONTAINER_STATUS_TO_WORKER_STATUS.get(container_status)
        if worker_status is None:
            return SyftError(message=f"Unknown container status: {container_status}")

        if isinstance(worker_status, SyftError):
            return worker_status

        if worker_status != WorkerStatus.PENDING:
            worker.status = worker_status

            result = self.stash.update(
                credentials=context.credentials,
                obj=worker_pool,
            )

            return (
                SyftError(
                    message=f"Failed to update worker status. Error: {result.err()}"
                )
                if result.is_err()
                else worker
            )

        return worker

    @service_method(
        path="worker_pool.get_worker_status",
        name="get_worker_status",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_worker_status(
        self, context: AuthedServiceContext, worker_pool_id: UID, worker_id: UID
    ) -> Union[WorkerStatus, SyftError]:
        worker = self.get_worker(context, worker_pool_id, worker_id)
        return worker if isinstance(worker, SyftError) else worker.status

    @service_method(
        path="worker_pool.worker_logs",
        name="worker_logs",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def worker_logs(
        self,
        context: AuthedServiceContext,
        worker_pool_id: UID,
        worker_id: UID,
        raw: bool = False,
    ) -> Union[bytes, str, SyftError]:
        worker_pool_worker = self._get_worker_pool_and_worker(
            context, worker_pool_id, worker_id
        )
        if isinstance(worker_pool_worker, SyftError):
            return worker_pool_worker

        _, worker = worker_pool_worker

        with contextlib.closing(docker.from_env()) as client:
            docker_container = _get_worker_container(client, worker)
            if isinstance(docker_container, SyftError):
                return docker_container

            try:
                logs = cast(bytes, docker_container.logs())
            except docker.errors.APIError as e:
                return SyftError(
                    f"Failed to get worker {worker.id} container logs. Error {e}"
                )

        return logs if raw else logs.decode(errors="ignore")

    def _get_worker_pool(
        self,
        context: AuthedServiceContext,
        worker_pool_id: UID,
    ) -> Union[WorkerPool, SyftError]:
        worker_pool = self.stash.get_by_uid(
            credentials=context.credentials, uid=worker_pool_id
        )

        return (
            SyftError(message=f"{worker_pool.err()}")
            if worker_pool.is_err()
            else cast(WorkerPool, worker_pool.ok())
        )

    def _get_worker_pool_and_worker(
        self, context: AuthedServiceContext, worker_pool_id: UID, worker_id: UID
    ) -> Union[Tuple[WorkerPool, SyftWorker], SyftError]:
        worker_pool = self._get_worker_pool(context, worker_pool_id)
        if isinstance(worker_pool, SyftError):
            return worker_pool

        worker = _get_worker(worker_pool, worker_id)
        if isinstance(worker, SyftError):
            return worker

        return worker_pool, worker


def _get_worker_opt(worker_pool: WorkerPool, worker_id: UID) -> Optional[SyftWorker]:
    try:
        return next(worker for worker in worker_pool.workers if worker.id == worker_id)
    except StopIteration:
        return None


def _get_worker(
    worker_pool: WorkerPool, worker_id: UID
) -> Union[SyftWorker, SyftError]:
    worker = _get_worker_opt(worker_pool, worker_id)
    return (
        worker
        if worker is not None
        else SyftError(
            message=f"Worker with id: {worker_id} not found in pool: {worker_pool.name}"
        )
    )


def _get_worker_container(
    client: docker.DockerClient,
    worker: SyftWorker,
) -> Union[Container, SyftError]:
    try:
        return cast(Container, client.containers.get(worker.container_id))
    except docker.errors.NotFound as e:
        return SyftError(f"Worker {worker.id} container not found. Error {e}")
    except docker.errors.APIError as e:
        return SyftError(
            f"Unable to access worker {worker.id} container. "
            + f"Container server error {e}"
        )


_CONTAINER_STATUS_TO_WORKER_STATUS: Dict[str, WorkerStatus] = dict(
    [
        ("running", WorkerStatus.RUNNING),
        *(
            (status, WorkerStatus.STOPPED)
            for status in ["paused", "removing", "exited", "dead"]
        ),
        ("restarting", WorkerStatus.RESTARTED),
        ("created", WorkerStatus.PENDING),
    ]
)


def _stop_worker_container(
    worker: SyftWorker,
    container: Container,
    force: bool,
) -> Optional[SyftError]:
    try:
        # stop the container
        container.stop()
        # Remove the container and its volumes
        _remove_worker_container(container, force=force, v=True)
    except Exception as e:
        return SyftError(
            message=f"Failed to delete worker with id: {worker.id}. Error: {e}"
        )


def _remove_worker_container(container: Container, **kwargs: Any) -> None:
    try:
        container.remove(**kwargs)
    except docker.errors.NotFound:
        return
    except docker.errors.APIError as e:
        if "removal of container" in str(e) and "is already in progress" in str(e):
            # If the container is already being removed, ignore the error
            return



# stdlib
from typing import Any

# relative
from ..service.response import SyftError


class InstallOrchestra:
    def launch(self, *args: Any, **kwargs: Any) -> None:
        return self.error()

    def error(self) -> Any:
        message = "Please install hagrid with `pip install -U hagrid`"
        return SyftError(message=message)

    def _repr_html_(self) -> str:
        return self.error()._repr_html_()


def import_orchestra() -> Any:
    try:
        # third party
        from hagrid import Orchestra

        return Orchestra

    except Exception as e:  # nosec
        print(e)
        pass
    return InstallOrchestra()


Orchestra = import_orchestra()



# stdlib
from enum import Enum
from typing import List
from typing import Optional

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...types.base import SyftBaseModel
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import short_uid
from ...types.uid import UID


@serializable()
class WorkerStatus(Enum):
    PENDING = "Pending"
    RUNNING = "Running"
    STOPPED = "Stopped"
    RESTARTED = "Restarted"


@serializable()
class WorkerHealth(Enum):
    HEALTHY = "✅"
    UNHEALTHY = "❌"


@serializable()
class SyftWorker(SyftObject):
    __canonical_name__ = "SyftWorker"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["name"]
    __attr_searchable__ = ["name", "container_id", "image_hash"]

    id: UID
    name: str
    container_id: Optional[str]
    created_at: DateTime = DateTime.now()
    image_hash: Optional[str]
    healthcheck: Optional[WorkerHealth]
    status: WorkerStatus
    job_id: Optional[UID]

    def get_job_repr(self):
        if self.job_id is not None:
            api = APIRegistry.api_for(
                node_uid=self.syft_node_location,
                user_verify_key=self.syft_client_verify_key,
            )
            job = api.services.job.get(self.job_id)
            if job.action.user_code_id is not None:
                func_name = api.services.code.get_by_id(
                    job.action.user_code_id
                ).service_func_name
                return f"{func_name} ({short_uid(self.job_id)})"
            else:
                return f"action ({short_uid(self.job_id)})"
        else:
            return ""

    def _coll_repr_(self):
        return {
            "name": self.name,
            "container id": self.container_id,
            "status": str(self.status.value.lower()),
            "job": self.get_job_repr(),
            "created at": str(self.created_at),
        }


@serializable()
class WorkerPool(SyftObject):
    __canonical_name__ = "WorkerPool"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["name"]
    __attr_searchable__ = ["name", "syft_worker_image_id"]

    name: str
    syft_worker_image_id: UID
    max_count: int
    workers: List[SyftWorker]


@serializable()
class WorkerOrchestrationType:
    DOCKER = "docker"
    K8s = "k8s"


@serializable()
class ContainerSpawnStatus(SyftBaseModel):
    __repr_attrs__ = ["worker_name", "worker", "error"]

    worker_name: str
    worker: Optional[SyftWorker]
    error: Optional[str]



# stdlib
from datetime import datetime
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

# third party
from IPython.display import HTML
from IPython.display import display
import itables
import pandas as pd
from pydantic import ValidationError
from pydantic import root_validator
from pydantic import validator
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...types.datetime import DateTime
from ...types.dicttuple import DictTuple
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.transforms import validate_url
from ...types.uid import UID
from ...util import options
from ...util.colors import ON_SURFACE_HIGHEST
from ...util.colors import SURFACE
from ...util.colors import SURFACE_SURFACE
from ...util.fonts import ITABLES_CSS
from ...util.fonts import fonts_css
from ...util.markdown import as_markdown_python_code
from ...util.notebook_ui.notebook_addons import FOLDER_ICON
from ...util.util import get_mb_size
from ..data_subject.data_subject import DataSubject
from ..data_subject.data_subject import DataSubjectCreate
from ..data_subject.data_subject_service import DataSubjectService
from ..response import SyftError
from ..response import SyftException
from ..response import SyftSuccess

DATA_SIZE_WARNING_LIMIT = 512


NamePartitionKey = PartitionKey(key="name", type_=str)


@serializable()
class Contributor(SyftObject):
    __canonical_name__ = "Contributor"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str
    role: Optional[str]
    email: str
    phone: Optional[str]
    note: Optional[str]

    __repr_attrs__ = ["name", "role", "email"]

    def _repr_html_(self) -> Any:
        return f"""
            <style>
            .syft-contributor {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='syft-contributor' style="line-height:25%">
                <h3>Contributor</h3>
                <p><strong>Name: </strong>{self.name}</p>
                <p><strong>Role: </strong>{self.role}</p>
                <p><strong>Email: </strong>{self.email}</p>
            </div>
            """

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Contributor):
            return False

        # We assoctiate two contributors as equal if they have the same email
        return self.email == value.email

    def __hash__(self) -> int:
        return hash(self.email)


@serializable()
class MarkdownDescription(SyftObject):
    # version
    __canonical_name__ = "MarkdownDescription"
    __version__ = SYFT_OBJECT_VERSION_1

    text: str

    def _repr_markdown_(self):
        style = """
        <style>
            .jp-RenderedHTMLCommon pre {
                background-color: #282c34 !important;
                padding: 10px 10px 10px;
            }
            .jp-RenderedHTMLCommon pre code {
                background-color: #282c34 !important;  /* Set the background color for the text in the code block */
                color: #abb2bf !important;  /* Set text color */
            }
        </style>
        """
        display(HTML(style))
        return self.text


@serializable()
class Asset(SyftObject):
    # version
    __canonical_name__ = "Asset"
    __version__ = SYFT_OBJECT_VERSION_1

    action_id: UID
    node_uid: UID
    name: str
    description: Optional[MarkdownDescription] = None
    contributors: Set[Contributor] = set()
    data_subjects: List[DataSubject] = []
    mock_is_real: bool = False
    shape: Optional[Tuple]
    created_at: DateTime = DateTime.now()
    uploader: Optional[Contributor]

    __repr_attrs__ = ["name", "shape"]

    def __init__(
        self, description: Optional[Union[MarkdownDescription, str]] = "", **data
    ):
        if isinstance(description, str):
            description = MarkdownDescription(text=description)
        super().__init__(**data, description=description)

    def _repr_html_(self) -> Any:
        itables_css = f"""
        .itables table {{
            margin: 0 auto;
            float: left;
            color: {ON_SURFACE_HIGHEST[options.color_theme]};
        }}
        .itables table th {{color: {SURFACE_SURFACE[options.color_theme]};}}
        """

        # relative
        from ...service.action.action_object import ActionObject

        uploaded_by_line = (
            f"<p><strong>Uploaded by: </strong>{self.uploader.name} ({self.uploader.email})</p>"
            if self.uploader
            else ""
        )

        if isinstance(self.data, ActionObject):
            data_table_line = itables.to_html_datatable(
                df=self.data.syft_action_data, css=itables_css
            )
        elif isinstance(self.data, pd.DataFrame):
            data_table_line = itables.to_html_datatable(df=self.data, css=itables_css)
        else:
            data_table_line = self.data

        if isinstance(self.mock, ActionObject):
            mock_table_line = itables.to_html_datatable(
                df=self.mock.syft_action_data, css=itables_css
            )
        elif isinstance(self.mock, pd.DataFrame):
            mock_table_line = itables.to_html_datatable(df=self.mock, css=itables_css)
        else:
            mock_table_line = self.mock
            if isinstance(mock_table_line, SyftError):
                mock_table_line = mock_table_line.message

        return f"""
            <style>
            {fonts_css}
            .syft-asset {{color: {SURFACE[options.color_theme]};}}
            .syft-asset h3,
            .syft-asset p
              {{font-family: 'Open Sans'}}
            {ITABLES_CSS}
            </style>

            <div class="syft-asset">
            <h3>{self.name}</h3>
            <p>{self.description}</p>
            <p><strong>Asset ID: </strong>{self.id}</p>
            <p><strong>Action Object ID: </strong>{self.action_id}</p>
            {uploaded_by_line}
            <p><strong>Created on: </strong>{self.created_at}</p>
            <p><strong>Data:</strong></p>
            {data_table_line}
            <p><strong>Mock Data:</strong></p>
            {mock_table_line}
            </div>"""

    def _repr_markdown_(self) -> str:
        _repr_str = f"Asset: {self.name}\n"
        _repr_str += f"Pointer Id: {self.action_id}\n"
        _repr_str += f"Description: {self.description}\n"
        _repr_str += f"Total Data Subjects: {len(self.data_subjects)}\n"
        _repr_str += f"Shape: {self.shape}\n"
        _repr_str += f"Contributors: {len(self.contributors)}\n"
        for contributor in self.contributors:
            _repr_str += f"\t{contributor.name}: {contributor.email}\n"
        return as_markdown_python_code(_repr_str)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Asset):
            return False
        return (
            self.action_id == other.action_id
            and self.name == other.name
            and self.contributors == other.contributors
            and self.shape == other.shape
            and self.description == other.description
            and self.data_subjects == other.data_subjects
            and self.mock_is_real == other.mock_is_real
            and self.uploader == other.uploader
            and self.created_at == other.created_at
        )

    @property
    def pointer(self) -> Any:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            node_uid=self.node_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        return api.services.action.get_pointer(self.action_id)

    @property
    def mock(self) -> Union[SyftError, Any]:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            node_uid=self.node_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        result = api.services.action.get_mock(self.action_id)
        try:
            if isinstance(result, SyftObject):
                return result.syft_action_data
            return result
        except Exception as e:
            return SyftError(message=f"Failed to get mock. {e}")

    def has_data_permission(self):
        return self.data is not None

    def has_permission(self, data_result):
        # TODO: implement in a better way
        return not (
            isinstance(data_result, str)
            and data_result.startswith("Permission")
            and data_result.endswith("denied")
        )

    @property
    def data(self) -> Any:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            node_uid=self.node_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        res = api.services.action.get(self.action_id)
        if self.has_permission(res):
            return res.syft_action_data
        else:
            return None


def _is_action_data_empty(obj: Any) -> bool:
    # just a wrapper of action_object.is_action_data_empty
    # to work around circular import error

    # relative
    from ...service.action.action_object import is_action_data_empty

    return is_action_data_empty(obj)


def check_mock(data: Any, mock: Any) -> bool:
    if type(data) == type(mock):
        return True

    return _is_action_data_empty(mock)


@serializable()
class CreateAsset(SyftObject):
    # version
    __canonical_name__ = "CreateAsset"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID] = None
    name: str
    description: Optional[MarkdownDescription] = None
    contributors: Set[Contributor] = set()
    data_subjects: List[DataSubjectCreate] = []
    node_uid: Optional[UID]
    action_id: Optional[UID]
    data: Optional[Any]
    mock: Optional[Any]
    shape: Optional[Tuple]
    mock_is_real: bool = False
    created_at: Optional[DateTime]
    uploader: Optional[Contributor]

    __repr_attrs__ = ["name"]

    class Config:
        validate_assignment = True

    def __init__(self, description: Optional[str] = "", **data):
        super().__init__(**data, description=MarkdownDescription(text=str(description)))

    @root_validator()
    def __empty_mock_cannot_be_real(cls, values: dict[str, Any]) -> Dict:
        """set mock_is_real to False whenever mock is None or empty"""

        if (mock := values.get("mock")) is None or _is_action_data_empty(mock):
            values["mock_is_real"] = False

        return values

    @validator("mock_is_real")
    def __mock_is_real_for_empty_mock_must_be_false(
        cls, v: bool, values: dict[str, Any], **kwargs: Any
    ) -> bool:
        if v and ((mock := values.get("mock")) is None or _is_action_data_empty(mock)):
            raise ValueError("mock_is_real must be False if mock is not provided")

        return v

    def add_data_subject(self, data_subject: DataSubject) -> None:
        self.data_subjects.append(data_subject)

    def add_contributor(
        self,
        name: str,
        email: str,
        role: Optional[Union[Enum, str]] = None,
        phone: Optional[str] = None,
        note: Optional[str] = None,
    ) -> Union[SyftSuccess, SyftError]:
        try:
            _role_str = role.value if isinstance(role, Enum) else role
            contributor = Contributor(
                name=name, role=_role_str, email=email, phone=phone, note=note
            )
            if contributor in self.contributors:
                return SyftError(
                    message=f"Contributor with email: '{email}' already exists in '{self.name}' Asset."
                )
            self.contributors.add(contributor)

            return SyftSuccess(
                message=f"Contributor '{name}' added to '{self.name}' Asset."
            )
        except Exception as e:
            return SyftError(message=f"Failed to add contributor. Error: {e}")

    def set_description(self, description: str) -> None:
        self.description = MarkdownDescription(text=description)

    def set_obj(self, data: Any) -> None:
        if isinstance(data, SyftError):
            raise SyftException(data)
        self.data = data

    def set_mock(self, mock_data: Any, mock_is_real: bool) -> None:
        if isinstance(mock_data, SyftError):
            raise SyftException(mock_data)

        current_mock = self.mock
        self.mock = mock_data

        try:
            self.mock_is_real = mock_is_real
        except ValidationError as e:
            self.mock = current_mock
            raise e

    def no_mock(self) -> None:
        # relative
        from ..action.action_object import ActionObject

        self.set_mock(ActionObject.empty(), False)

    def set_shape(self, shape: Tuple) -> None:
        self.shape = shape

    def check(self) -> Union[SyftSuccess, SyftError]:
        if not check_mock(self.data, self.mock):
            return SyftError(
                message=f"set_obj type {type(self.data)} must match set_mock type {type(self.mock)}"
            )
        if not _is_action_data_empty(self.mock):
            data_shape = get_shape_or_len(self.data)
            mock_shape = get_shape_or_len(self.mock)
            if data_shape != mock_shape:
                return SyftError(
                    message=f"set_obj shape {data_shape} must match set_mock shape {mock_shape}"
                )
        total_size_mb = get_mb_size(self.data) + get_mb_size(self.mock)
        if total_size_mb > DATA_SIZE_WARNING_LIMIT:
            print(
                f"**WARNING**: The total size for asset: '{self.name}' exceeds '{DATA_SIZE_WARNING_LIMIT} MB'. "
                "This might result in failure to upload dataset. "
                "Please contact #support on OpenMined slack for further assistance.",
            )

        return SyftSuccess(message="Dataset is Valid")


def get_shape_or_len(obj: Any) -> Optional[Union[Tuple[int, ...], int]]:
    if hasattr(obj, "shape"):
        shape = getattr(obj, "shape", None)
        if shape:
            return shape
    len_attr = getattr(obj, "__len__", None)
    if len_attr is not None:
        return len_attr()
    return None


@serializable()
class Dataset(SyftObject):
    # version
    __canonical_name__ = "Dataset"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    name: str
    node_uid: Optional[UID]
    asset_list: List[Asset] = []
    contributors: Set[Contributor] = set()
    citation: Optional[str]
    url: Optional[str]
    description: Optional[MarkdownDescription] = None
    updated_at: Optional[str]
    requests: Optional[int] = 0
    mb_size: Optional[int]
    created_at: DateTime = DateTime.now()
    uploader: Contributor

    __attr_searchable__ = ["name", "citation", "url", "description", "action_ids"]
    __attr_unique__ = ["name"]
    __repr_attrs__ = ["name", "url", "created_at"]

    def __init__(
        self, description: Optional[Union[str, MarkdownDescription]] = "", **data
    ):
        if isinstance(description, str):
            description = MarkdownDescription(text=description)
        super().__init__(**data, description=description)

    @property
    def icon(self):
        return FOLDER_ICON

    def _coll_repr_(self) -> Dict[str, Any]:
        return {
            "Name": self.name,
            "Assets": len(self.asset_list),
            "Size": f"{self.mb_size} (MB)",
            "Url": self.url,
            "created at": str(self.created_at),
        }

    def _repr_html_(self) -> Any:
        uploaded_by_line = (
            (
                "<p class='paragraph-sm'><strong>"
                + f"<span class='pr-8'>Uploaded by:</span></strong>{self.uploader.name} ({self.uploader.email})</p>"
            )
            if self.uploader
            else ""
        )

        return f"""
            <style>
            {fonts_css}
            .syft-dataset {{color: {SURFACE[options.color_theme]};}}
            .syft-dataset h3,
            .syft-dataset p
              {{font-family: 'Open Sans';}}
              {ITABLES_CSS}
            </style>
            <div class='syft-dataset'>
            <h3>{self.name}</h3>
            <p>{self.description.text}</p>
            {uploaded_by_line}
            <p class='paragraph-sm'><strong><span class='pr-8'>Created on: </span></strong>{self.created_at}</p>
            <p class='paragraph-sm'><strong><span class='pr-8'>URL:
            </span></strong><a href='{self.url}'>{self.url}</a></p>
            <p class='paragraph-sm'><strong><span class='pr-8'>Contributors:</span></strong>
            to see full details call <strong>dataset.contributors</strong></p>
            {self.assets._repr_html_()}
            """

    def action_ids(self) -> List[UID]:
        data = []
        for asset in self.asset_list:
            if asset.action_id:
                data.append(asset.action_id)
        return data

    @property
    def assets(self) -> DictTuple[str, Asset]:
        return DictTuple((asset.name, asset) for asset in self.asset_list)

    def _old_repr_markdown_(self) -> str:
        _repr_str = f"Syft Dataset: {self.name}\n"
        _repr_str += "Assets:\n"
        for asset in self.asset_list:
            _repr_str += f"\t{asset.name}: {asset.description.text}\n"
        if self.citation:
            _repr_str += f"Citation: {self.citation}\n"
        if self.url:
            _repr_str += f"URL: {self.url}\n"
        if self.description:
            _repr_str += f"Description: {self.description.text}\n"
        return as_markdown_python_code(_repr_str)

    def _repr_markdown_(self) -> str:
        # return self._old_repr_markdown_()
        return self._markdown_()

    def _markdown_(self) -> str:
        _repr_str = f"Syft Dataset: {self.name}\n\n"
        _repr_str += "Assets:\n\n"
        for asset in self.asset_list:
            _repr_str += f"\t{asset.name}: {asset.description.text}\n\n"
        if self.citation:
            _repr_str += f"Citation: {self.citation}\n\n"
        if self.url:
            _repr_str += f"URL: {self.url}\n\n"
        if self.description:
            _repr_str += f"Description: \n\n{self.description.text}\n\n"
        return _repr_str

    @property
    def client(self) -> Optional[Any]:
        # relative
        from ...client.client import SyftClientSessionCache

        client = SyftClientSessionCache.get_client_for_node_uid(self.node_uid)
        if client is None:
            return SyftError(
                message=f"No clients for {self.node_uid} in memory. Please login with sy.login"
            )
        return client


_ASSET_WITH_NONE_MOCK_ERROR_MESSAGE: str = "".join(
    [
        "To be included in a Dataset, an asset must either contain a mock, ",
        "or have it explicitly set to be empty.\n",
        "You can create an asset without a mock with `sy.Asset(..., mock=sy.ActionObject.empty())` or\n"
        "set the mock of an existing asset to be empty with `asset.no_mock()` or ",
        "`asset.mock = sy.ActionObject.empty()`.",
    ]
)


def _check_asset_must_contain_mock(asset_list: List[CreateAsset]) -> None:
    assets_without_mock = [asset.name for asset in asset_list if asset.mock is None]
    if assets_without_mock:
        raise ValueError(
            "".join(
                [
                    "These assets do not contain a mock:\n",
                    *[f"{asset}\n" for asset in assets_without_mock],
                    "\n",
                    _ASSET_WITH_NONE_MOCK_ERROR_MESSAGE,
                ]
            )
        )


@serializable()
class DatasetPageView(SyftObject):
    # version
    __canonical_name__ = "DatasetPageView"
    __version__ = SYFT_OBJECT_VERSION_1

    datasets: DictTuple
    total: int


@serializable()
class CreateDataset(Dataset):
    # version
    __canonical_name__ = "CreateDataset"
    __version__ = SYFT_OBJECT_VERSION_1
    asset_list: List[CreateAsset] = []

    __repr_attrs__ = ["name", "url"]

    id: Optional[UID] = None
    created_at: Optional[DateTime]
    uploader: Optional[Contributor]

    class Config:
        validate_assignment = True

    def _check_asset_must_contain_mock(self) -> None:
        _check_asset_must_contain_mock(self.asset_list)

    @validator("asset_list")
    def __assets_must_contain_mock(
        cls, asset_list: List[CreateAsset]
    ) -> List[CreateAsset]:
        _check_asset_must_contain_mock(asset_list)
        return asset_list

    def set_description(self, description: str) -> None:
        self.description = MarkdownDescription(text=description)

    def add_citation(self, citation: str) -> None:
        self.citation = citation

    def add_url(self, url: str) -> None:
        self.url = url

    def add_contributor(
        self,
        name: str,
        email: str,
        role: Optional[Union[Enum, str]] = None,
        phone: Optional[str] = None,
        note: Optional[str] = None,
    ) -> Union[SyftSuccess, SyftError]:
        try:
            _role_str = role.value if isinstance(role, Enum) else role
            contributor = Contributor(
                name=name, role=_role_str, email=email, phone=phone, note=note
            )
            if contributor in self.contributors:
                return SyftError(
                    message=f"Contributor with email: '{email}' already exists in '{self.name}' Dataset."
                )
            self.contributors.add(contributor)
            return SyftSuccess(
                message=f"Contributor '{name}' added to '{self.name}' Dataset."
            )
        except Exception as e:
            return SyftError(message=f"Failed to add contributor. Error: {e}")

    def add_asset(
        self, asset: CreateAsset, force_replace=False
    ) -> Union[SyftSuccess, SyftError]:
        if asset.mock is None:
            raise ValueError(_ASSET_WITH_NONE_MOCK_ERROR_MESSAGE)

        for i, existing_asset in enumerate(self.asset_list):
            if existing_asset.name == asset.name:
                if not force_replace:
                    return SyftError(
                        message=f"""Asset "{asset.name}" already exists in '{self.name}' Dataset."""
                        """ Use add_asset(asset, force_replace=True) to replace."""
                    )
                else:
                    self.asset_list[i] = asset
                    return SyftSuccess(
                        f"Asset {asset.name} has been successfully replaced."
                    )

        self.asset_list.append(asset)

        return SyftSuccess(
            message=f"Asset '{asset.name}' added to '{self.name}' Dataset."
        )

    def replace_asset(self, asset: CreateAsset):
        return self.add_asset(asset=asset, force_replace=True)

    def remove_asset(self, name: str) -> None:
        asset_to_remove = None
        for asset in self.asset_list:
            if asset.name == name:
                asset_to_remove = asset
                break

        if asset_to_remove is None:
            return SyftError(message=f"No asset exists with name: {name}")
        self.asset_list.remove(asset_to_remove)
        return SyftSuccess(
            message=f"Asset '{self.name}' removed from '{self.name}' Dataset."
        )

    def check(self) -> Result[SyftSuccess, List[SyftError]]:
        errors = []
        for asset in self.asset_list:
            result = asset.check()
            if not result:
                errors.append(result)
        if len(errors):
            return Err(errors)
        return Ok(SyftSuccess(message="Dataset is Valid"))


def create_and_store_twin(context: TransformContext) -> TransformContext:
    action_id = context.output["action_id"]
    if action_id is None:
        # relative
        from ...types.twin_object import TwinObject

        private_obj = context.output.pop("data", None)
        mock_obj = context.output.pop("mock", None)
        if private_obj is None and mock_obj is None:
            raise Exception("No data and no action_id means this asset has no data")

        twin = TwinObject(
            private_obj=private_obj,
            mock_obj=mock_obj,
        )
        action_service = context.node.get_service("actionservice")
        result = action_service.set(
            context=context.to_node_context(), action_object=twin
        )
        if result.is_err():
            raise Exception(f"Failed to create and store twin. {result}")

        context.output["action_id"] = twin.id
    else:
        private_obj = context.output.pop("data", None)
        mock_obj = context.output.pop("mock", None)
    return context


def infer_shape(context: TransformContext) -> TransformContext:
    if context.output["shape"] is None:
        if not _is_action_data_empty(context.obj.mock):
            context.output["shape"] = get_shape_or_len(context.obj.mock)
    return context


def set_data_subjects(context: TransformContext) -> TransformContext:
    data_subjects = context.output["data_subjects"]
    get_data_subject = context.node.get_service_method(DataSubjectService.get_by_name)

    resultant_data_subjects = []
    for data_subject in data_subjects:
        result = get_data_subject(context=context, name=data_subject.name)
        if isinstance(result, SyftError):
            return result
        resultant_data_subjects.append(result)
    context.output["data_subjects"] = resultant_data_subjects
    return context


def add_msg_creation_time(context: TransformContext) -> TransformContext:
    context.output["created_at"] = DateTime.now()
    return context


@transform(CreateAsset, Asset)
def createasset_to_asset() -> List[Callable]:
    return [
        generate_id,
        add_msg_creation_time,
        infer_shape,
        create_and_store_twin,
        set_data_subjects,
    ]


def convert_asset(context: TransformContext) -> TransformContext:
    assets = context.output.pop("asset_list", [])
    for idx, create_asset in enumerate(assets):
        asset_context = TransformContext.from_context(obj=create_asset, context=context)
        assets[idx] = create_asset.to(Asset, context=asset_context)
    context.output["asset_list"] = assets
    return context


def add_current_date(context: TransformContext) -> TransformContext:
    current_date = datetime.now()
    formatted_date = current_date.strftime("%b %d, %Y")
    context.output["updated_at"] = formatted_date
    return context


@transform(CreateDataset, Dataset)
def createdataset_to_dataset() -> List[Callable]:
    return [
        generate_id,
        add_msg_creation_time,
        validate_url,
        convert_asset,
        add_current_date,
    ]


class DatasetUpdate:
    pass



# stdlib
from typing import List
from typing import Union

# relative
from ...node.worker_settings import WorkerSettings
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..queue.queue_stash import ActionQueueItem
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .job_stash import Job
from .job_stash import JobStash
from .job_stash import JobStatus


@instrument
@serializable()
class JobService(AbstractService):
    store: DocumentStore
    stash: JobStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = JobStash(store=store)

    @service_method(
        path="job.get",
        name="get",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Job], SyftError]:
        res = self.stash.get_by_uid(context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            res = res.ok()
            return res

    @service_method(
        path="job.get_all",
        name="get_all",
    )
    def get_all(self, context: AuthedServiceContext) -> Union[List[Job], SyftError]:
        res = self.stash.get_all(context.credentials)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            res = res.ok()
            return res

    @service_method(
        path="job.restart",
        name="restart",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def restart(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        res = self.stash.get_by_uid(context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=res.err())

        job = res.ok()
        job.status = JobStatus.CREATED
        self.update(context=context, job=job)

        task_uid = UID()
        worker_settings = WorkerSettings.from_node(context.node)

        queue_item = ActionQueueItem(
            id=task_uid,
            node_uid=context.node.id,
            syft_client_verify_key=context.credentials,
            syft_node_location=context.node.id,
            job_id=job.id,
            worker_settings=worker_settings,
            args=[],
            kwargs={"action": job.action},
        )

        context.node.queue_stash.set_placeholder(context.credentials, queue_item)
        context.node.job_stash.set(context.credentials, job)
        log_service = context.node.get_service("logservice")
        result = log_service.restart(context, job.log_id)

        if result.is_err():
            return SyftError(message=str(result.err()))

        return SyftSuccess(message="Great Success!")

    @service_method(
        path="job.update",
        name="update",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def update(
        self, context: AuthedServiceContext, job: Job
    ) -> Union[SyftSuccess, SyftError]:
        res = self.stash.update(context.credentials, obj=job)
        if res.is_err():
            return SyftError(message=res.err())
        res = res.ok()
        return SyftSuccess(message="Great Success!")

    @service_method(
        path="job.kill",
        name="kill",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def kill(
        self, context: AuthedServiceContext, id: UID
    ) -> Union[SyftSuccess, SyftError]:
        res = self.stash.get_by_uid(context.credentials, uid=id)
        if res.is_err():
            return SyftError(message=res.err())

        job = res.ok()
        if job.job_pid is not None and job.status == JobStatus.PROCESSING:
            job.status = JobStatus.INTERRUPTED
            res = self.stash.update(context.credentials, obj=job)
            if res.is_err():
                return SyftError(message=res.err())
            return SyftSuccess(message="Job killed successfully!")
        else:
            return SyftError(
                message="Job is not running or isn't running in multiprocessing mode."
                "Killing threads is currently not supported"
            )

    @service_method(
        path="job.get_subjobs",
        name="get_subjobs",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_subjobs(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Job], SyftError]:
        res = self.stash.get_by_parent_id(context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            return res.ok()

    @service_method(
        path="job.get_active", name="get_active", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_active(self, context: AuthedServiceContext) -> Union[List[Job], SyftError]:
        res = self.stash.get_active(context.credentials)
        if res.is_err():
            return SyftError(message=res.err())
        return res.ok()



# stdlib
from typing import List
from typing import Tuple
from typing import Union

# third party
from pydantic.error_wrappers import ValidationError

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...store.document_store import QueryKeys
from ...types.uid import UID
from ..code.user_code import UserVerifyKeyPartitionKey
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from .action_graph import ActionGraphStore
from .action_graph import ExecutionStatus
from .action_graph import NodeActionData
from .action_graph import NodeActionDataUpdate
from .action_graph import NodeType
from .action_object import Action
from .action_object import ActionObject

ExecutionStatusPartitionKey = PartitionKey(key="status", type_=ExecutionStatus)


@serializable()
class ActionGraphService(AbstractService):
    store: ActionGraphStore

    def __init__(self, store: ActionGraphStore):
        self.store = store

    @service_method(path="graph.add_action", name="add_action")
    def add_action(
        self, context: AuthedServiceContext, action: Action
    ) -> Union[NodeActionData, SyftError]:
        # Create a node for the action
        input_uids, output_uid = self._extract_input_and_output_from_action(
            action=action
        )
        node = NodeActionData.from_action(
            action=action, credentials=context.credentials
        )

        result = self.store.set(
            credentials=context.credentials, node=node, parent_uids=input_uids
        )

        if result.is_err():
            return SyftError(message=result.err())

        action_node = result.ok()

        if action_node.is_mutagen:
            # updated non-mutated successor for all nodes between
            # node_id and nm_successor_id
            result = self.store.update_non_mutated_successor(
                node_id=action.remote_self.id,
                nm_successor_id=action_node.id,
                credentials=context.credentials,
            )
        else:
            # Create a node for the result object
            node = NodeActionData(
                id=output_uid,
                user_verify_key=context.credentials,
                type=NodeType.ACTION_OBJECT,
            )

            result = self.store.set(
                credentials=context.credentials,
                node=node,
                parent_uids=[action.id],
            )

        if result.is_err():
            return SyftError(message=result.err())

        result_node = result.ok()

        return action_node, result_node

    @service_method(path="graph.add_action_obj", name="add_action_obj")
    def add_action_obj(
        self, context: AuthedServiceContext, action_obj: ActionObject
    ) -> Union[NodeActionData, SyftError]:
        node = NodeActionData.from_action_obj(
            action_obj=action_obj, credentials=context.credentials
        )
        result = self.store.set(
            credentials=context.credentials,
            node=node,
        )
        if result.is_err():
            return SyftError(message=result.err())

        return result.ok()

    def _extract_input_and_output_from_action(self, action: Action) -> Tuple[UID]:
        input_uids = set()

        if action.remote_self is not None:
            input_uids.add(action.remote_self.id)

        for arg in action.args:
            input_uids.add(arg.id)

        for _, kwarg in action.kwargs.items():
            input_uids.add(kwarg.id)

        output_uid = action.result_id.id

        return input_uids, output_uid

    def get(
        self, uid: UID, context: AuthedServiceContext
    ) -> Union[NodeActionData, SyftError]:
        result = self.store.get(uid=uid, credentials=context.credentials)
        if result.is_err():
            return SyftError(message=result.err())
        return result.ok()

    def remove_node(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.store.delete(
            uid=uid,
            credentials=context.credentials,
        )
        if result.is_ok():
            return SyftSuccess(
                message=f"Successfully deleted node with uid: {uid} from the graph."
            )

        return SyftError(message=result.err())

    def get_all_nodes(self, context: AuthedServiceContext) -> Union[List, SyftError]:
        result = self.store.nodes(context.credentials)
        if result.is_ok():
            return result.ok()

        return SyftError(message="Failed to fetch nodes from the graph")

    def get_all_edges(self, context: AuthedServiceContext) -> Union[List, SyftError]:
        result = self.store.edges(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message="Failed to fetch nodes from the graph")

    def update(
        self,
        context: AuthedServiceContext,
        uid: UID,
        node_data: NodeActionDataUpdate,
    ) -> Union[NodeActionData, SyftError]:
        result = self.store.update(
            uid=uid, data=node_data, credentials=context.credentials
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    def update_action_status(
        self,
        context: AuthedServiceContext,
        action_id: UID,
        status: ExecutionStatus,
    ) -> Union[SyftSuccess, SyftError]:
        try:
            node_data = NodeActionDataUpdate(status=status)
        except ValidationError as e:
            return SyftError(message=f"ValidationError: {e}")
        result = self.store.update(
            uid=action_id, data=node_data, credentials=context.credentials
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    def get_by_action_status(
        self, context: AuthedServiceContext, status: ExecutionStatus
    ) -> Union[List[NodeActionData], SyftError]:
        qks = QueryKeys(qks=[ExecutionStatusPartitionKey.with_obj(status)])

        result = self.store.query(qks=qks, credentials=context.credentials)
        if result.is_ok():
            return result.ok()

        return SyftError(message=result.err())

    def get_by_verify_key(
        self, context: AuthedServiceContext, verify_key: SyftVerifyKey
    ) -> Union[List[NodeActionData], SyftError]:
        # TODO: Add a Query for Credentials as well,
        qks = QueryKeys(qks=[UserVerifyKeyPartitionKey.with_obj(verify_key)])

        result = self.store.query(qks=qks, credentials=context.credentials)
        if result.is_ok():
            return result.ok()

        return SyftError(message=result.err())



# stdlib
import socket
from typing import List
from typing import Union

# third party
import docker

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.document_store import SyftSuccess
from ...types.uid import UID
from ...util.telemetry import instrument
from ..service import AbstractService
from ..service import AuthedServiceContext
from ..service import SyftError
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .worker_pool import SyftWorker
from .worker_pool import WorkerStatus
from .worker_stash import WorkerStash

# def get_default_env_vars(context: AuthedServiceContext):
#     if context.node.runs_in_docker:
#         # get env vars from current environment
#         return dict(os.environ)
#     else:
#         # read env vars from .env file
#         env_path = f"{context.node.host_syft_location}/packages/grid/.env"
#         with open(env_path) as f:
#             lines = f.read().splitlines()

#         default_env_vars = {}
#         for line in lines:
#             if "=" in line:
#                 try:
#                     var_name, value = line.split("=", 1)

#                     def remove_redundant_quotes(value):
#                         for s in ['"', "'"]:
#                             if len(value) != 0:
#                                 if value[0] == s:
#                                     value = value[1:]
#                                 if value[-1] == s:
#                                     value = value[:-1]

#                     value = remove_redundant_quotes(value)
#                     default_env_vars[var_name] = value
#                 except Exception as e:
#                     print("error parsing env file", e)
#         return default_env_vars


# PORT_COUNTER = 0


# def get_env_vars(context: AuthedServiceContext):
#     default_env_vars = get_default_env_vars(context)
#     # stdlib
#     import secrets

#     worker_tag = "".join([str(secrets.choice(list(range(10)))) for i in range(10)])
#     node = context.node
#     # TODO, improve
#     global PORT_COUNTER
#     PORT_COUNTER += 1
#     extra_env_vars = {
#         "SERVICE_NAME": "backend",
#         "CREATE_PRODUCER": "false",
#         "N_CONSUMERS": "1",
#         "DEV_MODE": node.dev_mode,
#         "DEFAULT_ROOT_USERNAME": f"worker-{worker_tag}",
#         "PORT": str(8003 + PORT_COUNTER),
#         "QUEUE_PORT": node.queue_config.client_config.queue_port,
#         "HTTP_PORT": str(88 + PORT_COUNTER),
#         "HTTPS_PORT": str(446 + PORT_COUNTER),
#         "DEFAULT_ROOT_EMAIL": f"{worker_tag}@openmined.org",
#     }
#     # if node.dev_mode:
#     #     extra_env_vars["WATCHFILES_FORCE_POLLING"] = "true"

#     result = {**default_env_vars, **extra_env_vars}
#     result.pop("NODE_PRIVATE_KEY", None)
#     return result


def get_main_backend() -> str:
    hostname = socket.gethostname()
    return f"{hostname}-backend-1"


def start_worker_container(
    worker_num: int, context: AuthedServiceContext, syft_worker_uid
):
    client = docker.from_env()
    existing_container_name = get_main_backend()
    hostname = socket.gethostname()
    worker_name = f"{hostname}-worker-{worker_num}"
    return create_new_container_from_existing(
        worker_name=worker_name,
        client=client,
        existing_container_name=existing_container_name,
        syft_worker_uid=syft_worker_uid,
    )


def create_new_container_from_existing(
    worker_name: str,
    client: docker.client.DockerClient,
    existing_container_name: str,
    syft_worker_uid,
) -> docker.models.containers.Container:
    # Get the existing container
    existing_container = client.containers.get(existing_container_name)

    # Inspect the existing container
    details = existing_container.attrs

    # Extract relevant settings
    image = details["Config"]["Image"]
    command = details["Config"]["Cmd"]
    environment = details["Config"]["Env"]
    ports = details["NetworkSettings"]["Ports"]
    host_config = details["HostConfig"]

    volumes = {}
    for vol in host_config["Binds"]:
        parts = vol.split(":")
        key = parts[0]
        bind = parts[1]
        mode = parts[2]
        if "/storage" in bind:
            # we need this because otherwise we are using the same node private key
            # which will make account creation fail
            worker_postfix = worker_name.split("-", 1)[1]
            key = f"{key}-{worker_postfix}"
        volumes[key] = {"bind": bind, "mode": mode}

    # we need this because otherwise we are using the same node private key
    # which will make account creation fail

    environment = dict([e.split("=", 1) for e in environment])
    environment["CREATE_PRODUCER"] = "false"
    environment["N_CONSUMERS"] = 1
    environment["DOCKER_WORKER_NAME"] = worker_name
    environment["DEFAULT_ROOT_USERNAME"] = worker_name
    environment["DEFAULT_ROOT_EMAIL"] = f"{worker_name}@openmined.org"
    environment["PORT"] = str(8003 + WORKER_NUM)
    environment["HTTP_PORT"] = str(88 + WORKER_NUM)
    environment["HTTPS_PORT"] = str(446 + WORKER_NUM)
    environment["SYFT_WORKER_UID"] = str(syft_worker_uid)

    environment.pop("NODE_PRIVATE_KEY", None)

    new_container = client.containers.create(
        name=worker_name,
        image=image,
        command=command,
        environment=environment,
        ports=ports,
        detach=True,
        volumes=volumes,
        tty=True,
        stdin_open=True,
        network_mode=f"container:{existing_container.id}",
    )

    new_container.start()
    return new_container


WORKER_NUM = 0


@instrument
@serializable()
class WorkerService(AbstractService):
    store: DocumentStore
    stash: WorkerStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = WorkerStash(store=store)

    @service_method(
        path="worker.start_workers", name="start_workers", roles=ADMIN_ROLE_LEVEL
    )
    def start_workers(
        self, context: AuthedServiceContext, n: int = 1
    ) -> Union[SyftSuccess, SyftError]:
        """Add a Container Image."""
        for _worker_num in range(n):
            global WORKER_NUM
            WORKER_NUM += 1
            worker_uid = UID()
            container = start_worker_container(
                WORKER_NUM, context, syft_worker_uid=worker_uid
            )
            # worker_name = f"{pool_name}-{worker_count}"
            status = (
                WorkerStatus.STOPPED
                if container.status == "exited"
                else WorkerStatus.PENDING
            )
            obj = SyftWorker(
                id=worker_uid,
                name=f"default_pool-{WORKER_NUM}",
                container_id=container.id,
                image_hash=container.image.id,
                status=status,
            )
            # obj = SyftWorker(
            #     name: str
            #     container_id: str
            #     created_at: DateTime = DateTime.now()
            #     image_hash: str
            #     healthcheck: Optional[WorkerHealth]
            #     status: WorkerStatus
            # )
            # obj = DockerWorker(
            #     container_name=res.name, container_id=res.id, created_at=DateTime.now()
            # )
            result = self.stash.set(context.credentials, obj)
            if result.is_err():
                return SyftError(message=f"Failed to start worker. {result.err()}")

        return SyftSuccess(message=f"{n} workers added")

    @service_method(path="worker.list", name="list", roles=ADMIN_ROLE_LEVEL)
    def list(self, context: AuthedServiceContext) -> Union[SyftSuccess, SyftError]:
        """Add a Container Image."""
        result = self.stash.get_all(context.credentials)

        if result.is_err():
            return SyftError(message=f"Failed to fetch workers. {result.err()}")
        else:
            return result.ok()

    @service_method(path="worker.get", name="get", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        """Add a Container Image."""
        result = self.stash.get_by_uid(context.credentials, uid)

        if result.is_err():
            return SyftError(message=f"Failed to fetch worker. {result.err()}")
        else:
            return result.ok()

    @service_method(path="worker.stop", name="stop", roles=ADMIN_ROLE_LEVEL)
    def stop(
        self,
        context: AuthedServiceContext,
        workers: Union[List[SyftWorker], SyftWorker],
    ) -> Union[SyftSuccess, SyftError]:
        # listify
        if isinstance(workers, SyftWorker):
            workers = [workers]

        client = docker.from_env()
        for w in workers:
            result = self.stash.delete_by_uid(context.credentials, uid=w.id)

            if result.is_err():
                return SyftError(message=f"Failed to stop workers {result.err()}")

            # stop container
            try:
                client.containers.list(filters={"id": w.container_id})[0].stop()
                # also prune here?
            except Exception as e:
                # we dont throw an error here because apparently the container was already killed
                print(f"Failed to kill container {e}")

        return SyftSuccess(message=f"{len(workers)} workers stopped")



"""This module contains all the external libraries that Syft supports.
    We lazy load the external libraries when they are needed.
"""
# stdlib
import importlib
import os
from typing import Union

# relative
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..util.util import str_to_bool

# Contains all the external libraries that Syft supports.
# Used to check if a library is supported
# if the external library is not installed, we prompt the user
# to install it with the pip package name.

OBLV = str_to_bool(os.getenv("ENABLE_OBLV", "false"))

EXTERNAL_LIBS = {
    "oblv": {
        "pip_package_name": "oblv-ctl",
        "module_name": "oblv_ctl",
    }
}


def package_exists(package_name: str) -> bool:
    try:
        importlib.import_module(package_name)
        return True
    except ImportError:
        return False


def enable_external_lib(lib_name: str) -> Union[SyftSuccess, SyftError]:
    if lib_name in EXTERNAL_LIBS:
        syft_module_name = f"syft.external.{lib_name}"
        pip_package_name = EXTERNAL_LIBS[lib_name]["pip_package_name"]
        if not package_exists(EXTERNAL_LIBS[lib_name]["module_name"]):
            return SyftError(
                message=f"Package: {pip_package_name} for library: {lib_name} not installed.\n"
                + f"Kindly install it with 'pip install {pip_package_name}'"
            )

        importlib.import_module(syft_module_name)
        return SyftSuccess(message=f"Successfully enabled external library: {lib_name}")
    else:
        return SyftError(
            message=f"External library {lib_name} not supported. \n"
            + f"Supported external libraries are: {list(EXTERNAL_LIBS.keys())}"
        )



# stdlib
from typing import Dict
from typing import List
from typing import Optional

# third party
from typing_extensions import Self

# relative
from ..abstract_node import AbstractNode
from ..node.credentials import SyftVerifyKey
from ..node.credentials import UserLoginCredentials
from ..types.syft_object import Context
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.uid import UID
from .user.user_roles import ROLE_TO_CAPABILITIES
from .user.user_roles import ServiceRole
from .user.user_roles import ServiceRoleCapability


class NodeServiceContext(Context, SyftObject):
    __canonical_name__ = "NodeServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1
    id: Optional[UID]
    node: Optional[AbstractNode]


class AuthedServiceContext(NodeServiceContext):
    __canonical_name__ = "AuthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    credentials: SyftVerifyKey
    role: ServiceRole = ServiceRole.NONE
    job_id: Optional[UID]
    extra_kwargs: Dict = {}
    has_execute_permissions: bool = False

    def capabilities(self) -> List[ServiceRoleCapability]:
        return ROLE_TO_CAPABILITIES.get(self.role, [])

    def with_credentials(self, credentials: SyftVerifyKey, role: ServiceRole):
        return AuthedServiceContext(credentials=credentials, role=role, node=self.node)

    def as_root_context(self):
        return AuthedServiceContext(
            credentials=self.node.verify_key, role=ServiceRole.ADMIN, node=self.node
        )

    @property
    def job(self):
        if self.job_id is None:
            return None
        res = self.node.job_stash.get_by_uid(self.credentials, self.job_id)
        if res.is_err():
            return None
        else:
            return res.ok()


class UnauthedServiceContext(NodeServiceContext):
    __canonical_name__ = "UnauthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    login_credentials: UserLoginCredentials
    node: Optional[AbstractNode]
    role: ServiceRole = ServiceRole.NONE


class ChangeContext(SyftBaseObject):
    __canonical_name__ = "ChangeContext"
    __version__ = SYFT_OBJECT_VERSION_1

    node: Optional[AbstractNode] = None
    approving_user_credentials: Optional[SyftVerifyKey]
    requesting_user_credentials: Optional[SyftVerifyKey]
    extra_kwargs: Dict = {}

    @staticmethod
    def from_service(context: AuthedServiceContext) -> Self:
        return ChangeContext(
            node=context.node,
            approving_user_credentials=context.credentials,
            extra_kwargs=context.extra_kwargs,
        )



# stdlib
import os
from typing import Optional


def str_to_bool(bool_str: Optional[str]) -> bool:
    result = False
    bool_str = str(bool_str).lower()
    if bool_str == "true" or bool_str == "1":
        result = True
    return result


GEVENT_MONKEYPATCH = str_to_bool(os.environ.get("GEVENT_MONKEYPATCH", "False"))

# 🟡 TODO 30: Move this to where we manage the different concurrency modes later
# make sure its stable in containers and other run targets
# if GEVENT_MONKEYPATCH:
#     monkey.patch_all(ssl=False)


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


jupyter_notebook = is_notebook()



# stdlib
import asyncio
from enum import Enum
import logging
import multiprocessing
import os
import platform
import signal
import subprocess  # nosec
import time
from typing import Callable
from typing import List
from typing import Tuple

# third party
from fastapi import APIRouter
from fastapi import FastAPI
import requests
from starlette.middleware.cors import CORSMiddleware
import uvicorn

# relative
from ..abstract_node import NodeSideType
from ..client.client import API_PATH
from ..util.constants import DEFAULT_TIMEOUT
from ..util.util import os_name
from .domain import Domain
from .enclave import Enclave
from .gateway import Gateway
from .node import NodeType
from .routes import make_routes

if os_name() == "macOS":
    # needed on MacOS to prevent [__NSCFConstantString initialize] may have been in
    # progress in another thread when fork() was called.
    multiprocessing.set_start_method("spawn", True)

WAIT_TIME_SECONDS = 20


def make_app(name: str, router: APIRouter) -> FastAPI:
    app = FastAPI(
        title=name,
    )

    api_router = APIRouter()

    api_router.include_router(router)
    app.include_router(api_router, prefix="/api/v2")

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    return app


worker_classes = {
    NodeType.DOMAIN: Domain,
    NodeType.GATEWAY: Gateway,
    NodeType.ENCLAVE: Enclave,
}


def run_uvicorn(
    name: str,
    node_type: Enum,
    host: str,
    port: int,
    processes: int,
    reset: bool,
    dev_mode: bool,
    node_side_type: str,
    enable_warnings: bool,
):
    async def _run_uvicorn(
        name: str,
        node_type: Enum,
        host: str,
        port: int,
        reset: bool,
        dev_mode: bool,
        node_side_type: Enum,
    ):
        if node_type not in worker_classes:
            raise NotImplementedError(f"node_type: {node_type} is not supported")
        worker_class = worker_classes[node_type]
        if dev_mode:
            print(
                f"\nWARNING: private key is based on node name: {name} in dev_mode. "
                "Don't run this in production."
            )

            worker = worker_class.named(
                name=name,
                processes=processes,
                reset=reset,
                local_db=True,
                node_type=node_type,
                node_side_type=node_side_type,
                enable_warnings=enable_warnings,
                migrate=True,
            )
        else:
            worker = worker_class(
                name=name,
                processes=processes,
                local_db=True,
                node_type=node_type,
                node_side_type=node_side_type,
                enable_warnings=enable_warnings,
                migrate=True,
            )
        router = make_routes(worker=worker)
        app = make_app(worker.name, router=router)

        if reset:
            try:
                python_pids = find_python_processes_on_port(port)
                for pid in python_pids:
                    print(f"Stopping process on port: {port}")
                    kill_process(pid)
                    time.sleep(1)
            except Exception:  # nosec
                print(f"Failed to kill python process on port: {port}")

        log_level = "critical"
        if dev_mode:
            log_level = "info"
            logging.getLogger("uvicorn").setLevel(logging.CRITICAL)
            logging.getLogger("uvicorn.access").setLevel(logging.CRITICAL)
        config = uvicorn.Config(
            app, host=host, port=port, log_level=log_level, reload=dev_mode
        )
        server = uvicorn.Server(config)

        await server.serve()
        asyncio.get_running_loop().stop()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(
        _run_uvicorn(
            name,
            node_type,
            host,
            port,
            reset,
            dev_mode,
            node_side_type,
        )
    )
    loop.close()


def serve_node(
    name: str,
    node_type: NodeType = NodeType.DOMAIN,
    node_side_type: NodeSideType = NodeSideType.HIGH_SIDE,
    host: str = "0.0.0.0",  # nosec
    port: int = 8080,
    processes: int = 1,
    reset: bool = False,
    dev_mode: bool = False,
    tail: bool = False,
    enable_warnings: bool = False,
) -> Tuple[Callable, Callable]:
    server_process = multiprocessing.Process(
        target=run_uvicorn,
        args=(
            name,
            node_type,
            host,
            port,
            processes,
            reset,
            dev_mode,
            node_side_type,
            enable_warnings,
        ),
    )

    def stop():
        print(f"Stopping {name}")
        server_process.terminate()
        server_process.join(3)
        if server_process.is_alive():
            # this is needed because often the process is still alive
            server_process.kill()
            print("killed")

    def start():
        print(f"Starting {name} server on {host}:{port}")
        server_process.start()

        if tail:
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                try:
                    stop()
                except SystemExit:
                    os._exit(130)
        else:
            for i in range(WAIT_TIME_SECONDS):
                try:
                    req = requests.get(
                        f"http://{host}:{port}{API_PATH}/metadata",
                        timeout=DEFAULT_TIMEOUT,
                    )
                    if req.status_code == 200:
                        print(" Done.")
                        break
                except Exception:
                    time.sleep(1)
                    if i == 0:
                        print("Waiting for server to start", end="")
                    else:
                        print(".", end="")

    return start, stop


def find_python_processes_on_port(port: int) -> List[int]:
    system = platform.system()

    if system == "Windows":
        command = f"netstat -ano | findstr :{port}"
        process = subprocess.Popen(  # nosec
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        output, _ = process.communicate()
        pids = [
            int(line.strip().split()[-1]) for line in output.split("\n") if line.strip()
        ]

    else:  # Linux and MacOS
        command = f"lsof -i :{port} -sTCP:LISTEN -t"
        process = subprocess.Popen(  # nosec
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        output, _ = process.communicate()
        pids = [int(pid.strip()) for pid in output.split("\n") if pid.strip()]

    python_pids = []
    for pid in pids:
        try:
            if system == "Windows":
                command = (
                    f"wmic process where (ProcessId='{pid}') get ProcessId,CommandLine"
                )
            else:
                command = f"ps -p {pid} -o pid,command"

            process = subprocess.Popen(  # nosec
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            output, _ = process.communicate()
            lines = output.strip().split("\n")

            if len(lines) > 1 and "python" in lines[1].lower():
                python_pids.append(pid)

        except Exception as e:
            print(f"Error checking process {pid}: {e}")

    return python_pids


def kill_process(pid: int) -> None:
    try:
        os.kill(pid, signal.SIGTERM)
        print(f"Process {pid} terminated.")
    except Exception as e:
        print(f"Error killing process {pid}: {e}")



# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING

# third party
from typing_extensions import Self

# relative
from ..node.credentials import SyftVerifyKey
from ..serde.serializable import serializable
from .base import SyftBaseModel
from .uid import UID

if TYPE_CHECKING:
    # relative
    from ..client.client import SyftClient


class Identity(SyftBaseModel):
    node_id: UID
    verify_key: SyftVerifyKey

    __repr_attrs__ = ["id", "verify_key"]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <id={self.node_id.short()}, 🔑={str(self.verify_key)[0:8]}>"

    @classmethod
    def from_client(cls, client: SyftClient) -> Self:
        return cls(node_id=client.id, verify_key=client.credentials.verify_key)


@serializable()
class UserIdentity(Identity):
    """This class is used to identify the data scientist users of the node"""

    pass



# stdlib
from typing import Any
from typing import ClassVar
from typing import Optional
from typing import Type
from typing import Union

# relative
from ...serde.serializable import serializable
from ..response import SyftError
from ..response import SyftSuccess


@serializable()
class QueueClientConfig:
    pass


@serializable()
class AbstractMessageHandler:
    queue_name: ClassVar[str]

    @staticmethod
    def handle_message(message: bytes, consumer_id: str):
        raise NotImplementedError


@serializable(attrs=["message_handler", "queue_name", "address"])
class QueueConsumer:
    message_handler: AbstractMessageHandler
    queue_name: str
    address: str

    def receive(self):
        raise NotImplementedError

    def run(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError


@serializable()
class QueueProducer:
    address: str
    queue_name: str

    def send(
        self,
        message: Any,
    ):
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError


@serializable()
class QueueClient:
    pass


@serializable()
class QueueConfig:
    """Base Queue configuration"""

    client_type: Type[QueueClient]
    client_config: QueueClientConfig


@serializable()
class BaseQueueManager:
    config: QueueConfig

    def __init__(self, config: QueueConfig):
        self.config = config
        self.post_init()

    def post_init(self) -> None:
        pass

    def close(self) -> None:
        raise NotImplementedError

    def create_consumer(
        self,
        message_handler: Type[AbstractMessageHandler],
        address: Optional[str],
    ) -> QueueConsumer:
        raise NotImplementedError

    def create_producer(self, queue_name: str) -> QueueProducer:
        raise NotImplementedError

    def send(self, message: bytes, queue_name: str) -> Union[SyftSuccess, SyftError]:
        raise NotImplementedError

    @property
    def publisher(self) -> QueueProducer:
        raise NotImplementedError
